#
# VXLAN module
#
import typing
from box import Box
import netaddr

from . import _Module,get_effective_module_attribute,_dataplane
from .. import common
from .. import data
from ..data import get_from_box
from ..data.validate import must_be_int,must_be_string
from ..augment import devices
from .. import addressing

"""
register_static_vni -- register all static VNIs
"""
def register_static_vni(topology: Box) -> None:
  _dataplane.create_id_set('vni')
  _dataplane.extend_id_set('vni',_dataplane.build_id_set(topology,'vlans','vni','topology'))
  _dataplane.set_id_counter('vni',topology.defaults.vxlan.start_vni,16777215)

  for n in topology.nodes.values():
    _dataplane.extend_id_set('vni',_dataplane.build_id_set(n,'vlans','vni',f'nodes.{n.name}'))


"""
validate_vxlan_list

* Create vxlan.vlans list with all known VLANs if the attribute is missing
* validate that the vxlan.vlans list is a list with valid local or global VLAN names

The heavy lifting is done in a shared _dataplane function
"""
def validate_vxlan_list(toponode: Box, obj_path: str, topology: Box) -> None:
  _dataplane.validate_object_reference_list(
    parent=toponode if not toponode is topology else None,
    parent_path=obj_path,
    topology=topology, 
    list_name='vxlan.vlans',
    reference_dictionary='vlans',
    reference_name='VLAN',
    module='vxlan')

"""
assign_vni -- Assign VNIs to VLANs listed in vxlan.vlans (or all VLANs if vxlan.vlans is missing)

Inputs:
* toponode: Topology or node
* obj_path: object path (topology or nodes.x)
* topology: pointer to topology so we can access global VLANs
"""
def assign_vni(toponode: Box, obj_path: str, topology: Box) -> None:
  vxlan_vlans = get_from_box(toponode,'vxlan.vlans')
  if not vxlan_vlans:                                             # No VXLAN-enabled VLANs in the current data object ==> nothing to do
    return

  vni_ids = _dataplane.get_id_set('vni')
  for vname in vxlan_vlans:
    if not vname in toponode.get('vlans',{}):                     # Skip VXLAN-enabled VLANs that are not present on a node
      continue
    vlan_data = toponode.vlans[vname]
    vpath = f'{obj_path}.vlans.{vname}'
    if vname in topology.get('vlans',{}) and toponode != topology:
      if 'vni' in topology.vlans[vname] and 'vni' in vlan_data:
        common.error(
          f'Cannot define VXLAN VNI for a global VLAN {vname} within node {toponode.name} VLAN data',
          common.IncorrectType,
          'vxlan')
      continue

    if 'vni' in vlan_data:
      if vlan_data.vni is False:                                  # Explicit request not to assign a VNI
        continue
      elif vlan_data.vni is True:                                 # Explicit request to assign VNI, pass through
        pass
      else:                                                       # Otherwise check that VNI is an int
        must_be_int(
          parent=vlan_data,
          key='vni',
          path=vpath,
          min_value=2,
          max_value=16777215,
          module='vxlan')
        continue

    vni_default = topology.defaults.vxlan.start_vni + vlan_data.id  # ... try to build VNI from VLAN ID
    if not vni_default in vni_ids:                                  # Is the VNI free?
      vlan_data.vni = vni_default                                   # ... great, take it
      vni_ids.add(vni_default)                                      # ... and add it to the list of used VNIs
    else:                                                           # Too bad, we had such a great idea but it failed
      vlan_data.vni = _dataplane.get_next_id('vni')                 # ... so take the next available VNI

#
# Set VTEP IPv4/IPv6 address
#
def node_set_vtep(node: Box, topology: Box) -> bool:
  if topology.defaults.vxlan.use_v6_vtep and not 'ipv6' in node.loopback:
    common.error(
      f'You want to use IPv6 VTEP -- VXLAN module needs an IPv6 address on loopback interface of {node.name}',
      common.IncorrectValue,
      'vxlan')
    return False

  if not 'ipv4' in node.loopback and not topology.defaults.vxlan.use_v6_vtep:
    common.error(
      f'VXLAN module needs an IPv4 address on loopback interface of {node.name}',
      common.IncorrectValue,
      'vxlan')
    return False

  vtep_ip = ""
  vtep_af = 'ipv6' if topology.vxlan.use_v6_vtep else 'ipv4'
  vtep_ip = node.loopback[vtep_af]
  node.vxlan.vtep = str(netaddr.IPNetwork(vtep_ip).ip)              # ... and convert IPv4(v6) prefix into an IPv4(v6) address
  return True

#
# Build VLAN-specific VTEP list
#
def build_vtep_list(vlan: Box, node: str, nodes: typing.List[str], topology: Box) -> list:
  vlan.vtep_list = []                                               # Start with an emtpy VTEP list
  for n in nodes:
    if n == node:                                                   # Skip own node
      continue
    ndata = topology.nodes[n]
    if not 'vlans' in ndata:                                        # No VLANs on remote node, skip it
      continue
    vni_match = filter(lambda x: x.vni == vlan.vni,ndata.vlans.values())
    if list(vni_match):                                             # Is there a VLAN with matching VNI on remote node?
      vlan.vtep_list.append(ndata.vxlan.vtep)                       # ... if so, add remote VTEP to VLAN flood list

  return_value = vlan.vtep_list                                     # We'll return whatever we built
  if not vlan.vtep_list:                                            # ... but will remove empty VTEP list from VLAN data
    vlan.pop('vtep_list')

  return return_value

class VXLAN(_Module):

  def module_pre_transform(self, topology: Box) -> None:
    register_static_vni(topology)

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    flooding_values = ['static']
    for m in ['evpn']:
      if m in node.module:
        flooding_values.append(m)

    must_be_string(
      parent = node.vxlan,
      key = 'flooding',
      path = f'nodes.{node.name}.vxlan',
      valid_values = flooding_values)

  def module_post_node_transform(self, topology: Box) -> None:
    validate_vxlan_list(topology,'topology',topology)
    assign_vni(topology,'topology',topology)
    for n in topology.nodes.values():
      validate_vxlan_list(n,f'nodes.{n.name}',topology)
      assign_vni(n,f'nodes.{n.name}',topology)

  # We need multi-step post-transform node handling, so we have to do that in
  # module_post_transform
  #
  def module_post_transform(self, topology: Box) -> None:
    vxlan_domain_list: typing.Dict[str,list] = {}

    if not 'use_v6_vtep' in topology.vxlan:                         # Copy IPv6 VTEP setting into global parameter
      topology.vxlan.use_v6_vtep = topology.defaults.vxlan.use_v6_vtep

    for name,ndata in topology.nodes.items():
      if not 'vxlan' in ndata.get('module',[]):                     # Skip nodes without VXLAN module
        continue
      if not ndata.interfaces:
        common.error(
          f'VXLAN-enabled node {name} should be connected to at least one link',
          common.MissingValue,
          'vxlan')
        continue
      if not 'vlans' in ndata:                                      # Skip VXLAN-enabled nodes without VLANs
        if 'vxlan' in ndata:                                        # ... but make sure there's no vxlan.vlans list left on them
          ndata.vxlan.pop('vlans',None)
        continue

      # Build the final per-node VXLAN list -- all VLANs with VNI attribute are VXLAN-enabled in one way or another
      #
      ndata.vxlan.vlans = [ vname for vname,vdata in ndata.vlans.items() if 'vni' in vdata and vdata.vni ]
      if not node_set_vtep(ndata,topology):                         # Set VTEP IP address
        continue

      vxlan_domain = ndata.vxlan.domain                             # Find VXLAN flooding domain name
      if not vxlan_domain in vxlan_domain_list:                     # Unknown VXLAN domain, prepare an empty list
        vxlan_domain_list[vxlan_domain] = []
      vxlan_domain_list[vxlan_domain].append(name)                  # Add current node to VXLAN domain list

    for domain,nodes in vxlan_domain_list.items():                  # Iterate over VXLAN flooding domains
      for node in nodes:                                            # ... and over all nodes in each domain
        ndata = topology.nodes[node]
        if ndata.vxlan.get('flooding') != 'static':                 # Skip nodes that are not using static replication lists
          continue

        vtep_set = set()
        for vlan in ndata.vlans.values():                           # Iterate over all VLANs defined in current node
          if 'vni' in vlan:                                         # Are we dealing with VXLAN-enabled VLAN?
            vtep_list = build_vtep_list(vlan,node,nodes,topology)   # Build VLAN-specific VTEP list
            vtep_set.update(vtep_list)                              # ... and add it to node-level VTEP set


        ndata.vxlan.vtep_list = sorted(list(vtep_set))              # Convert node-level VTEP set into a list



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import subprocess

from box import Box

from .. import common
from .. import read_topology
from ..utils import status
from ..utils import strings

from . import external_commands
from . import down

def status_parse(args: typing.List[str]) -> argparse.Namespace:
  parser = argparse.ArgumentParser(
    prog='netlab status',
    description='Display lab status')
#  parser.add_argument(
#    '-v','--verbose',
#    dest='verbose',
#    action='count',
#    default=0,
#    help='Verbose logging')
  parser.add_argument(
    dest='action',
    action='store',
    choices=['list','show','cleanup','reset'],
    help='Lab status action')
  parser.add_argument(
    dest='instance',
    action='store',
    nargs="*",
    help='Display or cleanup specific lab instance(s)')
  parser.add_argument(
    '--all',
    dest='all',
    action='store_true',
    help='Display or cleanup all lab instance(s)')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='store_true',
    help='Verbose printout(s)')
  if args:
    return parser.parse_args(args)
  else:
    return parser.parse_args(['list'])

def display_active_labs(topology: Box,args: argparse.Namespace,lab_states: Box) -> None:
  if args.verbose:
    print(f'netlab status file: {status.get_status_filename(topology)}\n')
  if not lab_states:
    print('No netlab-managed labs')
    return

  for id,lab_state in lab_states.items():
    print(f'Lab {id} in {lab_state.dir}')
    print(f'  status: {lab_state.status}')
    print(f'  provider(s): {",".join(lab_state.providers)}')
    print()

  for provider,pdata in topology.defaults.providers.items():
    if not 'act_probe' in pdata:
      continue
    try:
      result = subprocess.run(pdata.act_probe.split(' '),capture_output=True,text=True)
    except:
      continue
    if result.stdout:
      print(f'{pdata.act_title}\n{"=" * 80}\n{result.stdout}\n')

def show_lab_instance(iid: str, lab_state: Box) -> None:
  print(f'Lab {iid} in {lab_state.dir}')
  print(f'  status: {lab_state.status}')
  print(f'  provider(s): {",".join(lab_state.providers)}')
  if lab_state.log:
    newline = "\n"
    print(f'\nLog:\n{"=" * 80}\n{newline.join(lab_state.log)}{newline}')

def show_lab(topology: Box,args: argparse.Namespace,lab_states: Box) -> None:
  for iid in args.instance:
    if not iid in lab_states:
      print(f'Unknown lab instance {iid}, skipping\n')
      continue

  for iid in lab_states.keys():
    if not iid in args.instance and not args.all:
      continue
    show_lab_instance(iid,lab_states[iid])

def cleanup_lab(topology: Box,args: argparse.Namespace,lab_states: Box) -> None:
  if args.all:
    print(f'Active lab instances:\n')
    for k,v in lab_states.items():
      print (f'  {k} in {v.dir}')
    print("")
  
  if not args.all and not args.instance:
    print('No lab instances specified, nothing to do')
    return
  confirm_msg = f'Cleanup will remove all {"" if args.all else "specified "}lab instances. Are you sure?'
  if not strings.confirm(confirm_msg):
    return

  pwd = os.getcwd()
  for iid in list(lab_states.keys()):
    if not args.all and not iid in args.instance:
      continue
    print(f'Shutting down lab {iid} in {lab_states[iid].dir}')
    os.chdir(lab_states[iid].dir)
    try:
      result = subprocess.run(['netlab','down','--cleanup'],capture_output=False,check=True)
    except Exception as ex:
      print(f'Error shutting down lab {iid}: {ex}')
      print(f'... aborting the cleanup process')
      break

  os.chdir(pwd)

def reset_lab_status(topology: Box,args: argparse.Namespace,lab_states: Box) -> None:
  lab_status_file = status.get_status_filename(topology)
  print(f'''
This action deletes the lab status file {lab_status_file}.

That makes it impossible for netlab to track active lab instances. Use this
command only when you're absolutely sure that the lab status shown by
'netlab status' is corrupted
''')
  if not strings.confirm('Do you want to continue?'):
    return
  try:
    os.remove(lab_status_file)
    print('Lab status file removed')
  except Exception as ex:
    common.fatal(f'Cannot remove lab status file: {ex}')

action_map = {
  'list': display_active_labs,
  'show': show_lab,
  'cleanup': cleanup_lab,
  'reset': reset_lab_status
}

def run(cli_args: typing.List[str]) -> None:
  topology = read_topology.load("package:cli/empty.yml","","package:topology-defaults.yml")
  lab_states = status.read_status(topology)
  args = status_parse(cli_args)
  common.set_logging_flags(args)
  if args.action in action_map:
    action_map[args.action](topology,args,lab_states)
  else:
    common.fatal(f'Unknown action {args.action}')
    return



#
# Data validation routines
#

import typing,typing_extensions
import builtins as _bi
from box import Box
from .. import common
from . import get_from_box,set_dots,get_empty_box,get_box

#
# Import functions from data.types to cope with legacy calls to must_be_something
from .types import must_be_list,must_be_dict,must_be_string,must_be_int,must_be_bool

# We also need to import the whole data.types module to be able to do validation function lookup
from . import types as _tv

"""
get_valid_attributes

Given an attribute dictionary, list of valid attribute categories, and extra attributes, return a list
of valid attributes, or a string (type name) if the first attribute source in the list is a string
"""

##### REMOVE AFTER ATTRIBUTE MIGRATION #####
def make_attr_dict(atlist: typing.Union[list,Box]) -> typing.Union[Box]:
  if isinstance(atlist,list):
    return Box({ k: None for k in atlist })

  return atlist

def get_valid_attributes(
      attributes: Box,                                      # Where to get valid attributes from
      attr_list: typing.List[str],                          # List of valid attributes (example: ['node'] or ['link','interface'])
      extra_attributes: typing.Optional[list] = None        # List of dynamic attributes (needed to validate node provider settings)
        ) -> typing.Union[str,Box]:

  valid = get_empty_box()
  for idx,atlist in enumerate(attr_list):                   # Build a list of all valid (global) attributes for the object
    if not atlist in attributes:
      continue
    add_attr = attributes[atlist]                           # Attributes to add to the list

    if isinstance(add_attr,str):                            # Got a specific data type?
      if valid:                                             # Have we already collected something?
        common.fatal(
          f'Internal error trying to build list of attributes for {attr_list} -- unexpected value at {atlist}\n' +
          f'... attributes: {attributes}')
        return ''                                           # ... bad karma, inconsistent validation requirements
      return add_attr

    if not isinstance(add_attr,(Box,list)):
      common.fatal(
        f'Internal error: Expected string or list/dictionary for {atlist} attributes\n' +
        f'... attributes: {attributes}')
      return ''                                             # ... dang, someone messed up. Abort, abort, abort...

    add_attr = make_attr_dict(add_attr)                     # Convert attributes into a dictionary

    no_propagate = f'{atlist}_no_propagate'                 # No-propagate list excluded only for non-first attribute category
    if idx and no_propagate in attributes:
      add_attr = { k:v for k,v in add_attr.items() if not k in attributes[no_propagate] }
    valid += add_attr                                       # ... nope, add to list of attributes and move on

    internal_atlist = f'{atlist}_internal'                  # Internal object attributes (used by links)
    if internal_atlist in attributes:                       # Add internal attributes if they exist
      valid += make_attr_dict(attributes[internal_atlist])

  if not extra_attributes is None:                      # Extend the attribute list with dynamic attributes
    valid += make_attr_dict(extra_attributes)

  return valid

"""
build_module_extra_attributes -- build a list of extra attributes defined by modules

Some modules (vlan, vrf) define extra attributes (vlans, vrfs). While those attributes get added to the correct
attribute lists during module initialization phase, it's good to know what hasn't been enabled if those attributes
are not valid, so we're building a reverse lookup list from defaults._module_.attributes.extra lists
"""

def build_module_extra_attributes(topology: Box) -> None:
  global_attr = topology.defaults.attributes            # Shortcut to global attribute list (where 'extra' list will be added)
  # Iterate over all modules
  for m in topology.defaults.keys():                    # Iterate over all default settings
    if not 'supported_on' in topology.defaults[m]:      # ... not a module, move on
      continue
    mod_attr = topology.defaults[m].attributes          # A convenient shortcut to current module attributes to keep code simple
    if not 'extra' in mod_attr:                         # No extra attributes for this module, move on
      continue

    for a_cat,a_list in mod_attr.extra.items():         # Now iterate over all attribute categories defined by this module
      for attr in a_list:                               # ... and for every entry in that list
        global_attr.extra[a_cat][attr] = m              # ... add a pointer back to the module

"""
get_extra_module_attributes -- return a dictionary of extra module attributes relevant to current validation context
"""
def get_extra_module_attributes(attributes : Box,attr_list: typing.List[str]) -> dict:
  return { 
    attr:mod for a_cat,a_extra in attributes.extra.items() \
               if a_cat in attr_list \
                 for attr,mod in a_extra.items() }

"""
validate_module_can_be_false: Check whether module attributes for an object can be 'false'
"""
def validate_module_can_be_false(
      attributes: Box,                                  # Attribute definition
      attr_list: typing.List[str]                       # List of valid attributes (example: ['node'] or ['link','interface'])
        ) -> bool:

  if not 'can_be_false' in attributes:
    return False

  intersect = set(attr_list) & set(attributes.can_be_false)
  return bool(intersect)

"""
type_has_attribute: checks whether the data type definition has the specified attribute
"""

def type_has_attribute(data_type: typing.Any, attr: str) -> bool:
  return isinstance(data_type,Box) and attr in data_type

"""
validate_dictionary -- recursively validate a dictionary

Arguments are described in validate_item
"""

def validate_dictionary(
      data: Box,
      data_type: typing.Any,
      parent_path: str,
      data_name: str,
      module: str,
        ) -> bool:

  OK = True
  for k in data.keys():                                             # Iterate over the elements
    if not k in data_type:                                          # ... and report elements with invalid name
      common.error(
        f'Incorrect {data_name} attribute {k} in {parent_path}',
        common.IncorrectAttr,
        module)
      OK = False
    else:                                                           # For valid elements, validate them
      validate_item(
        parent=data,
        key=k,
        data_type=data_type[k],
        parent_path=parent_path,
        data_name=data_name,
        module=module)
  return OK

"""
validate_alt_type -- deal with dictionaries that could be specified as something else
"""

def validate_alt_type(data: typing.Any, data_type: Box) -> bool:
  return type(data).__name__ in data_type._alt_types

"""
validate_value -- validate a single value (not from an object)
"""

def validate_value(
      value: typing.Any,
      data_type: str,
      path: str,
      context: typing.Optional[typing.Any] = None,      # Additional context (use when verifying link values)
      module: typing.Optional[str] = None,              # Module name to display in error messages
      abort: bool = False) -> typing.Any:
  global _bi,_tv

  validation_function = getattr(_tv,f'must_be_{data_type}',None)      # Try to get validation function

  if not validation_function:                                         # No validation function
    if type(value).__name__ != data_type:                             # ... check data type name
      _tv.wrong_type_message(
              path=path,
              key=None,
              expected=data_type,
              value=value,
              context=context,
              module=module)
      if abort:
        raise common.IncorrectType()
    return value

  return validation_function(
            parent=None,                                # We're validating a standalone value
            key=value,
            path=path,
            module=module,
            context=context,
            abort=abort)

"""
validate_item -- validate a single item from an object:

* Return if the data type is None (= not validated)
* Compare data types names if the data type is a string (OK, a bit more complex than that)
* Recursively validate a dictionary

To make matters worse, we cannot pass the item-to-validate directly into the function
but have to invoke it with parent dictionary and key, so we can forward these elements
to "must_be_something" routines.
"""

def validate_item(
      parent: Box,
      key: str,
      data_type: typing.Any,
      parent_path: str,
      data_name: str,
      module: str,
        ) -> typing.Any:
  global _bi,_tv

  data = parent[key]
  if data_type is None:                                               # Trivial case - data type not specified
    return True                                                       # ==> anything goes

  # We have to handle a weird corner case: AF (or similar) list that is really meant to be a dictionary
  #
  if isinstance(data,list) and type_has_attribute(data_type,'_list_to_dict'):
    parent[key] = { k: data_type._list_to_dict for k in data }        # Transform lists into a dictionary (updating parent will make it into a Box)
    data = parent[key]
    data_type = Box(data_type)                                        # and fix datatype definition

  if isinstance(data,Box) and isinstance(data_type,Box):              # Validating a dictionary against a dictionary of elements
    return validate_dictionary(
              data=data,
              data_type=data_type,
              parent_path=f"{parent_path}.{key}",
              data_name=data_name,
              module=module)

  if type_has_attribute(data_type,'_alt_types'):
    if validate_alt_type(data,data_type):
      return True

  if isinstance(data_type,str):                                       # Convert desired data type name into a dummy data type dictionary
    data_type = { 'type': data_type }

  if isinstance(data_type,list):                                      # Convert list into 'list' datatype with 'valid_values'
    data_type = {
      'type': 'list',
      'valid_values': data_type
    }

  if not 'type' in data_type:                                         # The required data type is a true dict, but the data is not
    common.error(
      f'{data_name} attribute {parent_path}.{key} should be a dictionary, found {type(data).__name__}',
      common.IncorrectType,
      module)
    return False

  for kw in list(data_type.keys()):                                   # Remove all 'internal' type attributes before calling validation function
    if kw.startswith('_'):
      data_type.pop(kw)

  dt_name = data_type['type']                                         # Get the desired data type in string format
  validation_function = getattr(_tv,f'must_be_{dt_name}',None)        # Try to get validation function

  if not validation_function:                                         # No validation function, have to compare type names
    if not hasattr(_bi,dt_name):                                      # Is the requested data type a well-known type/class?
      common.fatal(f'Invalid data type {dt_name} found when trying to validate {data_name} attribute {parent_path}.{key}')
      return False                                                    # pragma: no cover

    dt_ref = getattr(_bi,dt_name)                                     # Get pointer to desired data type
    if not isinstance(data,dt_ref):                                   # ... and check if the current object is an instance of it
      common.error(
        f'{data_name} attribute {parent_path}.{key} should be {dt_name}, found {type(data).__name__}',
        common.IncorrectType,
        module)
      return False

    return True                                                       # We couldn't do full validation, but at least the instance type is OK

  # We have to validate an item with a validation function
  #
  # First, create the extra arguments for the must_be_something function:
  #
  # * Start with the data type definition converted into a dictionary
  # * Remove all attributes that are not used by the must_be_something functions
  #
  validation_attr = data_type.to_dict() if isinstance(data_type,Box) else data_type
  for kw in ['type']:
    validation_attr.pop(kw,None)

  if data_type in ('dict','list') and not 'create_empty' in validation_attr:
    validation_attr['create_empty'] = False                           # Do not create empty dictionaries/lists unless told otherwise

  # Now call the validation function and hope for the best ;)
  #
  return validation_function(
            parent=parent,                                            # We're validating a single item
            key=key,                                                  # So we're setting the retrieval key to None
            path=parent_path,                                         # Pass the parent path (it will be combined with key anyway)
            module=module,                                            # ... and the module
            **validation_attr)                                        # And any other attributes

"""
validate_attributes -- validate object attributes

Iterate over all keys in the 'data' dictionary and check whether they're valid global attributes
or module names. For module attributes, iterate over all valid module attributes

Returns the original data or None transformed into an empty dictionary
"""

list_of_modules: typing.List[str] = []

def validate_attributes(
      data: Box,                                        # Object to validate
      topology: Box,                                    # Pointer to topology
      data_path: str,                                   # Path to the data object (needed in error messages)
      data_name: str,                                   # Name of the object (needed in error messages, example: 'node')
      attr_list: typing.List[str],                      # List of valid attributes (example: ['node'] or ['link','interface'])
      modules: typing.Optional[list] = [],              # List of relevant modules
      module: str = 'attributes',                       # Module generating the error message (default: 'attributes')
      module_source: typing.Optional[str] = None,       # Where did we get the list of modules?
      attributes: typing.Optional[Box] = None,          # Where to get valid attributes from
      extra_attributes: typing.Optional[list] = None    # List of dynamic attributes (needed to validate node provider settings)
        ) -> typing.Any: 

  #
  # Part 1: set up default values
  #
  global list_of_modules

  if attributes is None:
    attributes = topology.defaults.attributes

  if not 'extra' in topology.defaults.attributes:
    build_module_extra_attributes(topology)

  if module_source is None:
    module_source = data_path

  if not list_of_modules:
    list_of_modules = [ m for m in topology.defaults.keys() if 'supported_on' in topology.defaults[m] ]

  if common.debug_active('validate'):
    print(f'validate {data_path} against {attr_list} attributes + {modules} modules from {module_source}')

  #
  # Part 2: Build the list of valid attributes
  #
  # It could be that the list of attributes tells us data should be of certain type
  # Deal with that as well (although in an awkward way that should be improved)
  #
  valid = get_valid_attributes(attributes,attr_list,extra_attributes)
  extra_module_attr = get_extra_module_attributes(topology.defaults.attributes,attr_list)
  if isinstance(valid,str):                   # Validate data that is not a dictionary
    validate_value(                           # Use standalone value validator
      value=data,
      data_type=valid,
      path=data_path,
      module=module)
    return data

  #
  # Now that we know the valid attributes are a list, add extra attributes from enabled modules
  #

  if not modules is None:
    valid += Box({ attr: None for attr,mod in extra_module_attr.items() if mod in modules })

  # Part 3 -- validate attributes
  #
  # * Anything starting with "_" is OK (internal attributes)
  # * Known attributes for object we're checking are OK
  # * Module attributes are OK, but have to be checked recursively
  # * Attributes from modules not used by the current object are NOT OK
  # * Anything else is clearly an error
  #
  if data is None:
    return {}                                           # Can't validate attributes of None, but maybe the caller can fix this
  if not isinstance(data,Box):
    common.error(
      f'Cannot validate attributes in {data_path} -- that should have been a dictionary, found {type(data).__name__}',
      common.IncorrectType,
      'validate')
    return

  for k in data.keys():
    if k.startswith('_'):                               # Skip internal attributes
      continue

    if k in valid:
      validate_item(
        parent=data,
        key=k,
        data_type=valid[k],
        parent_path=data_path,
        data_name=data_name,
        module=module)
      continue

    if not modules is None and k in modules:            # For module attributes, perform recursive check
      if data[k] is False and validate_module_can_be_false(attributes,attr_list):
        continue                                        # Some objects accept 'attribute: false' syntax (example: links)
      if data[k] is True:
        if set(attr_list) & set(topology.defaults[k].attributes.can_be_true):
          continue
      fixed_data = validate_attributes(
        data=data[k],
        topology=topology,
        data_path=f'{data_path}.{k}',                   # Change 'node' to 'node.bgp'
        data_name=f'{k} {data_name}',                   # Change 'node' to 'bgp node' to be used in 'bgp node attribute...'
        attr_list=attr_list,                            # Not changing the checking context
        modules=[],                                     # No extra modules to add
        module=k,                                       # Error message generated by the module
        attributes=topology.defaults[k].attributes)     # Use module attributes

      if data[k] is None:                               # If we're trying to validate module value None...
        data[k] = fixed_data                            # ... replace it with whatever recursive call returned (might be a dict)

      continue

    if k in list_of_modules and not modules is None:
      common.error(
        f"{data_path} uses an attribute from module {k} which is not enabled in {module_source}",
        common.IncorrectAttr,
        module)
      continue

    if k in extra_module_attr and not modules is None:
      common.error(
        f"Attribute '{k}' used in {data_path} is defined by module {extra_module_attr[k]} which is not enabled",
        common.IncorrectAttr,
        module)
      continue

    common.error(
      f"Invalid {data_name} attribute '{k}' found in {data_path}",
      common.IncorrectAttr,
      module)



'''
Create detailed node-level data structures from topology

* Discover desired imagex (boxes)
* Add default module list to nodes without specific modules
* Set loopback and management interface data
'''
import typing

from box import Box

from .. import common
from .. import utils
from .. import addressing
from .. import providers
from . import devices
from ..data.validate import validate_attributes
from ..data.types import must_be_int,must_be_string,must_be_id
from ..modules._dataplane import extend_id_set,is_id_used,set_id_counter,get_next_id

"""
Reserve a node ID, for example for gateway ID, return True if successful, False if duplicate
"""
def reserve_id(n_id: int) -> bool:
  if is_id_used('node_id',n_id):
    return False

  extend_id_set('node_id',set([n_id]))
  return True

"""
Node data structure is a dictionary. Convert lists of dictionaries (now obsolete)
Or lists of strings into a unified dictionary structure
"""

def create_node_dict(nodes: Box) -> Box:
  if isinstance(nodes,dict):
    node_dict = nodes
  else:
    node_dict = Box({},default_box=True,box_dots=True)
    for n in nodes or []:
      if isinstance(n,dict):
        if not 'name' in n:
          common.error(f'Node is missing a "name" attribute: {n}',common.IncorrectValue,'nodes')
          continue
      elif isinstance(n,str):
        n = Box({ 'name': n },default_box=True,box_dots=True)
      node_dict[n.name] = n

  for name in list(node_dict.keys()):
    ndata = node_dict[name]
    if ndata is None:
      ndata = Box({'name': name},default_box=True)
    elif not isinstance(ndata,dict):
      common.error(f'Node data for node {name} must be a dictionary')
      node_dict[name] = { 'name': name, 'extra': ndata }
      ndata = node_dict[name]
    else:
      ndata['name'] = name

    ndata.interfaces = ndata.interfaces or []   # Make sure node.interfaces is always defined
    node_dict[name] = ndata

  common.exit_on_error()
  return node_dict

"""
Validate node attributes
"""
def validate(topology: Box) -> None:
  for n_name,n_data in topology.nodes.items():
    must_be_id(parent=None,key=n_name,path=f'NOATTR:node name {n_name}',module='nodes')
    providers = list(topology.defaults.providers.keys())
    validate_attributes(
      data=n_data,                                    # Validate node data
      topology=topology,
      data_path=f'nodes.{n_name}',                    # Topology path to node name
      data_name=f'node',
      attr_list=['node'],                             # We're checking node attributes
      modules=n_data.get('module',[]),                # ... against node modules
      module='nodes',                                 # Function is called from 'nodes' module
      extra_attributes = providers)                   # Allow provider-specific settings (not checked at the moment)

def augment_mgmt_if(node: Box, defaults: Box, addrs: typing.Optional[Box]) -> None:
  if 'ifname' not in node.mgmt:
    mgmt_if = devices.get_device_attribute(node,'mgmt_if',defaults)
    if not mgmt_if:
      ifname_format = devices.get_device_attribute(node,'interface_name',defaults)
      if not isinstance(ifname_format,str):
        common.fatal("Missing interface name template for device type %s" % node.device)
        return

      ifindex_offset = devices.get_device_attribute(node,'ifindex_offset',defaults)
      if ifindex_offset is None:
        ifindex_offset = 1

      mgmt_if = utils.strings.eval_format(ifname_format,{'ifindex': ifindex_offset - 1 })

    node.mgmt.ifname = mgmt_if

  # If the mgmt ipaddress is statically set (IPv4/IPv6)
  # skip the address set
  if 'ipv4' in node.mgmt or 'ipv6' in node.mgmt:
    return

  if not addrs:                                                       # We need a management address pool, but there's none
    common.error(
      f"Node {node.name} does not have a management IP address and there's no 'mgmt' address pool",
      common.MissingValue,
      'nodes')
    return

  for af in 'ipv4','ipv6':                                            # Try to assign IPv4 or IPv6 management address
    pfx = af + '_pfx'
    if not pfx in addrs:                                              # ... desired AF not in management pool, try the next one
      continue

    if not addrs.get('start'):
      common.fatal("Start offset missing in management address pool for AF %s" % af)

    try:                                                              # Try to assign management address (might fail due to large ID)
      node.mgmt[af] = str(addrs[pfx][node.id+addrs.start])
    except Exception as ex:
      common.error(
        f'Cannot assign management address from prefix {str(addrs[pfx])} (starting at {addrs.start}) to node with ID {node.id}',
        common.IncorrectValue,
        'nodes')

  if addrs.mac_eui and not 'mac' in node.mgmt:                        # Finally, assign management MAC address
    addrs.mac_eui[5] = node.id
    node.mgmt.mac = str(addrs.mac_eui)

  if not 'ipv4' in node.mgmt and not 'ipv6' in node.mgmt:             # Final check: did we get a usable management address?
    common.error(
      f'Node {node.name} does not have a usable management IP addresss',
      common.MissingValue,
      'nodes')

"""
Add device data to nodes
"""

def find_node_device(n: Box, topology: Box) -> bool:
  if 'device' not in n:
    n.device = topology.defaults.device

  if not n.device:
    common.error(
      f'No device type specified for node {n.name} and there is no default device type',
      common.MissingValue,
      'nodes')
    return False

  try:
    must_be_string(n,'device',f'nodes.{n.name}',module='nodes',abort=True)
  except Exception as ex:
    return False

  devtype = n.device

  if not devtype in topology.defaults.devices:
    common.error(f'Unknown device {devtype} in node {n.name}',common.IncorrectValue,'nodes')
    return False

  dev_def = topology.defaults.devices[devtype]
  if not isinstance(dev_def,dict):
    common.fatal(f"Device data for device {devtype} must be a dictionary")

  for kw in ['interface_name','description']:
    if not kw in dev_def:
      common.error(
        f'Device {devtype} used on node {n.name} is not a valid device type\n'+
        "... run 'netlab show devices' to display valid device types",
        common.IncorrectValue,
        'nodes')
      return False

  return True

"""
Find the image/box for the container/device
"""
def find_node_image(n: Box, topology: Box) -> bool:
  provider = devices.get_provider(n,topology.defaults)

  pdata = devices.get_provider_data(n,topology.defaults)
  if 'node' in pdata:
    if not isinstance(pdata.node,Box):    # pragma: no cover
      common.fatal(f"Node data for device {n.device} provider {provider} must be a dictionary")
      return False
    n[provider] = pdata.node + n.get(provider,{})

  if n.box:
    return True

  if 'image' in n:
    n.box = n.image
    del n['image']
    return True

  if 'image' in topology.defaults.devices[n.device]:
    if not must_be_string(topology.defaults.devices[n.device],'image',f'defaults.devices.{n.device}',module='nodes'):
      return False

    n.box = topology.defaults.devices[n.device].image
    return True

  if 'image' in pdata:
    if not must_be_string(pdata,'image',f'defaults.devices.{n.device}.{provider}.image',module='nodes'):
      return False

    n.box = pdata.image
    return True

  common.error(
    f'No image specified for device {n.device} (provider {provider}) used by node {n.name}',
    common.MissingValue,
    'nodes')

  return False

"""
Validate provider setting used in a node
"""

def validate_node_provider(n: Box, topology: Box) -> bool:
  if not 'provider' in n:
    return True

  if not n.provider in topology.defaults.providers:
    common.error(
      f'Invalid provider {n.provider} specified in node {n.name}',
      common.IncorrectValue,
      'nodes')
    return False

  if not n.provider in topology.defaults.providers[topology.provider]:
    common.error(
      f'Provider {n.provider} specified in node {n.name} is not compatible with lab topology provider {topology.provider}',
      common.IncorrectValue,
      'nodes')
    return False

  topology[topology.provider].providers[n.provider] = True
  return True

"""
Add provider data to nodes:

* Check whether the node device exists
* Copy device.provider.node into node.provider
* Get device image
"""
def augment_node_provider_data(topology: Box) -> None:
  if not topology.defaults.devices:
    common.fatal('Device defaults (defaults.devices) are missing')

  for name,n in topology.nodes.items():
    if not validate_node_provider(n,topology):
      continue

    if not find_node_device(n,topology):
      continue

    if not find_node_image(n,topology):
      continue

"""
Add system data to devices -- hacks that are not yet covered in the settings structure
"""
def augment_node_system_data(topology: Box) -> None:
  if 'mtu' in topology.defaults.get('interfaces',{}):
    if not isinstance(topology.defaults.interfaces.mtu,int):            # pragma: no cover
      common.error(
        'defaults.interfaces.mtu setting should be an integer',
        common.IncorrectValue,
        'topology')
    else:
      for n in topology.nodes.values():
        if not 'mtu' in n:
          n.mtu = topology.defaults.interfaces.mtu
        else:
          if not isinstance(n.mtu,int):                                 # pragma: no cover
            common.error(
              f'nodes.{n.name}.mtu setting should be an integer',
              common.IncorrectValue,
              'nodes')

"""
augment_node_device_data: copy attributes that happen to be node attributes from device defaults into node data
"""

def augment_node_device_data(n: Box, defaults: Box) -> None:
  node_attr = defaults.attributes.get('node',[])
  dev_data  = devices.get_consolidated_device_data(n,defaults)

  for attr in node_attr:                      # Copy known node attributes from device+provider data into node data
    if attr in dev_data and not attr in n:
      n[attr] = dev_data[attr]

  if 'node' in defaults.devices[n.device]:    # Copy everything within device (but not provider) node dictionary into node data
    for k in defaults.devices[n.device].node.keys():
      if not k in n:
        n[k] = defaults.devices[n.device].node[k]

'''
Main node transformation code

* set node ID
* set loopback address(es)
* copy device data from defaults
* set management IP and MAC addresses
'''
def transform(topology: Box, defaults: Box, pools: Box) -> None:
  for name,n in topology.nodes.items():
    if not must_be_int(n,'id',f'nodes.{name}',module='nodes',min_value=1,max_value=250):
      continue
    if not reserve_id(n.id):
      common.error(
        f'Duplicate static node ID {n.id} on node {n.name}',
        common.IncorrectValue,
        'nodes')

  common.exit_on_error()
  set_id_counter('node_id',1,250)
  for name,n in topology.nodes.items():
    if not 'id' in n:
      n.id = get_next_id('node_id')

    if not n.name: # pragma: no cover (name should have been checked way before)
      common.fatal(f"Internal error: node does not have a name {n}",'nodes')
      return

    augment_node_device_data(n,defaults)

    if pools.loopback and n.get('role','') != 'host':
      prefix_list = addressing.get(pools,['loopback'],n.id)
      for af in prefix_list:
        if isinstance(prefix_list[af],bool):
          if prefix_list[af]:
            common.fatal( f"Loopback addresses must be valid IP prefixes, not 'True': {prefix_list}" )
        elif not n.loopback[af]:
          if af == 'ipv6':
            n.loopback[af] = addressing.get_addr_mask(prefix_list[af],1)
          else:
            n.loopback[af] = str(prefix_list[af])

    augment_mgmt_if(n,defaults,topology.addressing.mgmt)
    providers.execute_node("augment_node_data",n,topology)

'''
Return a copy of the topology (leaving original topology unchanged) with unmanaged devices removed
'''
def ghost_buster(topology: Box) -> Box:
  common.print_verbose('Removing unmanaged devices from topology')
  # Create a copy of topology
  topo_copy = Box(topology,default_box=True,box_dots=True)
  
  # Remove all nodes with "unmanaged" flag set
  topo_copy.nodes = { k:v for k,v in topo_copy.nodes.items() if not v.get('unmanaged',False) }  
  return topo_copy



#
# Implement global variables hidden within topology defaults
#
# Needed to support reentrant use of netsim modules (custom transformations or tests)
#

import typing
from box import Box

_topology: typing.Optional[Box] = None
_globals:  typing.Optional[Box] = None

'''
init -- create 'globals' entry in 'topology.defaults' and save a pointer to it

It would be ideal to get a pointer to topology in every call to 'get', but
that's not realistic.
'''

def init(topology: Box) -> None:
  global _topology,_globals

  _topology = topology
  _globals  = topology.defaults._globals


'''
get -- get a pointer to a global Box (referenced by name) hidden in topology
'''

def get(varname: str) -> Box:
  if _globals is None:
    from ..common import fatal

    fatal(f'Trying to get global variable {varname} before the global_vars subsystem is initialized')
    return Box({})                  # pragma: no cover -- fatal aborts, but we need to return the right object to keep mypy happy

  return _globals[varname]

def get_topology() -> typing.Optional[Box]:
  global _topology

  return _topology



#
# Common routines used to implement lab status and locking commands
#

import typing
import os
import sys
from box import Box
from filelock import Timeout, FileLock

from ..common import fatal, get_yaml_string,debug_active
from ..data import get_empty_box

'''
get_status_filename -- get the name of the netlab status file
'''
def get_status_filename(topology: Box) -> str:
  status_file = topology.defaults.lab_status_file or '~/.netlab/status.yaml'
  return os.path.expanduser(status_file)

'''
change_status -- change the status of a lab

* Lock the lab status file
* Read the YAML document in the lab status file
* Call a callback function to change the status
* Write the modified YAML document
* Unlock the lab status file
'''
def change_status(topology: Box, callback: typing.Callable[[Box,Box], None]) -> None:
  status_file = get_status_filename(topology)               # Get status file name from topology defaults
  lock_file   = f'{status_file}.lock'                       # Associated lock file

  try:
    status_dir = os.path.dirname(status_file)
    if not os.path.exists(status_dir):
      os.makedirs(status_dir)
  except:
    fatal(f'Cannot create lab status directory {status_dir}')

  try:                                                      # Try to lock the status file          
    lock = FileLock(lock_file, timeout=3)
    lock.acquire()
    if os.path.exists(status_file):                       # If the status file exists, read it
      try:
        status = Box().from_yaml(filename=status_file,default_box=True,box_dots=True)
      except:
        fatal(f'Cannot read lab status file {status_file}')
    else:                                                 # Otherwise, create an empty status
      status = get_empty_box()

    callback(status,topology)                               # Change the lab status
    if debug_active('status'):
      print(f'Lab status: {status}')
    with open(status_file, 'w') as f:                       # Write the modified status          
      f.write(get_yaml_string(status))
  except:
    fatal(f'Cannot lock lab status file {lock_file}\n... {sys.exc_info()[0]}')
  finally:
    lock.release()

def read_status(topology: Box) -> Box:
  status_file = get_status_filename(topology)               # Get status file name from topology defaults
  if not os.path.exists(status_file):
    return get_empty_box()
  
  try:
    return Box().from_yaml(filename=status_file,default_box=True,box_dots=True)
  except:
    fatal(f'Cannot read lab status file {status_file}')
    return get_empty_box()

lock_file: typing.Final[str] = 'netlab.lock'

'''
lock_directory -- create netlab.lock file in current directory to prevent 
                  overwriting provider configuration files or Ansible inventory
'''
def lock_directory() -> None:
  global lock_file
  with open(lock_file, 'w') as f:
    f.write('netlab lock file, do not remove')

'''
unlock_directory -- remove netlab.lock file in current directory
'''
def unlock_directory() -> None:
  global lock_file
  if os.path.exists(lock_file):
    os.remove(lock_file)

'''
is_locked -- check if netlab.lock file exists in current directory
'''
def is_directory_locked() -> bool:
  global lock_file
  return os.path.exists(lock_file)



#
# netlab up command
#
# * Transform lab topology and create provider and automation files,
#   or read transformed lab topology from snapshot file
# * Start the lab, including provider-specific pre- and post-start hooks
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box
from pathlib import Path

from .. import common
from . import create
from . import external_commands
from . import common_parse_args, get_message
from . import lab_status_update, lab_status_change, get_lab_id
from .. import providers
from .. import read_topology
from ..utils import status
from ..data import get_box,get_from_box

#
# Extra arguments for 'netlab up' command
#
def up_parse_args(standalone: bool) -> argparse.ArgumentParser:
  parse_parents = [ common_parse_args() ] if standalone else []
  parser = argparse.ArgumentParser(
    description='netlab up extra arguments',
    parents = parse_parents,
    add_help=standalone)
  parser.add_argument(
    '--no-config',
    dest='no_config',
    action='store_true',
    help='Do not configure lab devices')
  parser.add_argument(
    '--fast-config',
    dest='fast_config',
    action='store_true',
    help='Use fast device configuration (Ansible strategy = free)')
  parser.add_argument(
    '--snapshot',
    dest='snapshot',
    action='store',
    nargs='?',
    const='netlab.snapshot.yml',
    help='Use netlab snapshot file created by a previous lab run')
  return parser

"""
Get lab topology from the snapshot file or run the transformation process
"""
def get_topology(args: argparse.Namespace, cli_args: typing.List[str]) -> Box:
  up_args_parser = up_parse_args(bool(args.snapshot))         # (re)create the correct parser

  if args.snapshot:                                           # If we're using the snapshot file...
    args = up_args_parser.parse_args(cli_args)                # ... and reparse
    common.set_logging_flags(args)                            # ... use these arguments to set logging flags and read the snapshot

    topology = read_topology.read_yaml(filename=args.snapshot)
    if topology is None:
      common.fatal(f'Cannot read snapshot file {args.snapshot}, aborting...')
      return Box({})                                          # pragma: no-cover

    print(f"Using transformed lab topology from snapshot file {args.snapshot}")
  else:                                                       # No snapshot file, use 'netlab create' parser
    topology = create.run(cli_args,'up','Create configuration files, start a virtual lab, and configure it',up_args_parser)

  return topology

"""
Lab status routines:

* status_start_lab -- lab initialization has started
* status_start_provider -- provider activation has started
* status_config -- configuration deployment has started
* status_complete -- lab initialization has completed
"""

def lab_status_start(status: Box, topology: Box) -> None:
  lab_id = get_lab_id(topology)                             # Get the lab ID (or default)
  if lab_id in status:
    if status[lab_id].dir != os.getcwd():
      lab_status_update(topology,status,
        update = { 'status': f'conflict -- trying to start lab in {os.getcwd()}'})
      topology.defaults.err_conflict = status[lab_id].dir
      return

  lab_status_update(topology,status,
    update = {
      'status': 'starting lab' if not lab_id in status else 'restarting lab',
      'name': topology.name,
      'providers': [] })

def status_start_lab(topology: Box) -> None:
  status.change_status(
    topology,
    callback = lambda s,t: lab_status_start(s,t))

def status_start_provider(topology: Box, provider: str) -> None:
  status.change_status(
    topology,
    callback = lambda s,t: 
      lab_status_update(t,s,
        update = { 'status': f'starting provider {provider}' },
        cb = lambda s: s.providers.append(provider)))

"""
check_existing_lab -- print an command-specific error message if there'a s lab already running in this directory
"""
def check_existing_lab() -> None:
  if not status.is_directory_locked():
    return
  
  print(f'''
It looks like you have another lab running in this directory. If you want to
continue the lab startup process due to a previous failure, please use the
'netlab up --snapshot' command.

Otherwise use 'netlab status' to check the status of labs running on this machine, or
'netlab down' to shut down the other lab running in this directory.

If you are sure that no other lab is running in this directory, remove the
netlab.lock file manually and retry.
''')
  common.fatal('Cannot start another lab in the same directory')

"""
check_lab_instance -- print an error message if the lab instance is already running in a different directory
"""
def check_lab_instance(topology: Box) -> None:
  lab_id = get_lab_id(topology)                   # Get the current lab instance ID from lab topology
  lab_states = status.read_status(topology)       # Read the state of existing lab instances

  if not lab_id in lab_states:                    # If this lab instance is not running ==> OK
    return
  
  if lab_states[lab_id].dir == os.getcwd():       # If this lab instance is already running in this directory
    return                                        # ... we'll deal with that a bit later in the process

  print(f'''
It looks like the lab instance '{lab_id}' is already running in directory
{lab_states[lab_id].dir}.

Please use 'netlab status' to check the status of labs running on this machine.
You can stop the other lab instance with 'netlab status cleanup {lab_id}'.

If you think your netlab status file is corrupt, use 'netlab status reset' to
delete it.
''')
  common.fatal(f'aborting "netlab up" request')

"""
Execute provider probes
"""
def provider_probes(topology: Box) -> None:
  p_provider = topology.provider

  external_commands.run_probes(topology.defaults,p_provider,2)
  for s_provider in topology[p_provider].providers:
    external_commands.run_probes(topology.defaults,s_provider,2)

"""
Start lab topology for a single provider
"""
def start_provider_lab(topology: Box, pname: str, sname: typing.Optional[str] = None) -> None:
  p_name   = sname or pname
  p_module = providers._Provider.load(p_name,topology.defaults.providers[p_name])

  if sname is not None:
    p_topology = providers.select_topology(topology,p_name)
  else:
    p_topology = topology

  status_start_provider(topology,p_name)
  p_module.call('pre_start_lab',p_topology)
  if sname is not None:
    exec_command = topology.defaults.providers[pname][sname].start
  else:
    exec_command = topology.defaults.providers[pname].start

  exec_list = exec_command if isinstance(exec_command,list) else [ exec_command ]
  for cmd in exec_list:
    external_commands.start_lab(topology.defaults,sname or pname,3,"netlab up",cmd)
  p_module.call('post_start_lab',p_topology)

  lab_status_change(topology,f'{p_name} workload started')

"""
Recreate secondary configuration file
"""
def recreate_secondary_config(topology: Box, p_provider: str, s_provider: str) -> None:
  sp_data = topology.defaults.providers[p_provider][s_provider]
  if not sp_data.recreate_config:                                     # Do we need to recreate the config file?
    return

  sp_module  = providers._Provider.load(s_provider,topology.defaults.providers[s_provider])
  s_topology = providers.select_topology(topology,s_provider)         # Create secondary provider subtopology
  filename = sp_data.filename                                         # Get the secondary configuration filename
  print(f"Recreating {filename} configuration file for {s_provider} provider")
  sp_module.create(s_topology,filename)                               # ... and create the new configuration file

"""
Main "lab start" process
"""
def run(cli_args: typing.List[str]) -> None:
  up_args_parser = up_parse_args(False)                       # Try to parse the up-specific arguments
  (args,rest) = up_args_parser.parse_known_args(cli_args)
  if not args.snapshot:
    check_existing_lab()

  topology = get_topology(args,cli_args)
  check_lab_instance(topology)

  settings = topology.defaults
  if common.QUIET:
    os.environ["ANSIBLE_STDOUT_CALLBACK"] = "selective"

  provider_probes(topology)

  p_provider = topology.provider
  p_module = providers._Provider.load(p_provider,topology.defaults.providers[p_provider])
  providers.mark_providers(topology)
  p_module.call('pre_output_transform',topology)

  status_start_lab(topology)
  if 'err_conflict' in topology.defaults:
    common.fatal(f'race condition, lab instance already running in {topology.defaults.err_conflict}')
    return

  status.lock_directory()

  start_provider_lab(topology,p_provider)
  for s_provider in topology[p_provider].providers:
    print()
    recreate_secondary_config(topology,p_provider,s_provider)
    start_provider_lab(topology,p_provider,s_provider)

  if not args.no_config:
    lab_status_change(topology,f'deploying initial configuration')
    external_commands.deploy_configs(4,"netlab up",args.fast_config)
    message = get_message(topology,'up',False)
    if message:
      print(f"\n\n{message}")
    lab_status_change(topology,f'initial configuration complete')
  else:
    print("\nInitial configuration skipped, run 'netlab initial' to configure the devices")

  lab_status_change(topology,'started')



#
# Dynamic output framework
# 
# Individual output routines are defined in modules within this directory inheriting
# TopologyOutput class and replacing or augmenting its methods (most commonly, write)
#

import platform
import subprocess
import os
import typing
import re

# Related modules
from box import Box

from .. import common
from ..utils import status
from ..callback import Callback

'''
check_writeable -- check if the current directory is writeable (does not have netlab.lock file)

Print an error message explaining what to do next and exit with a fatal error if the directory is locked
'''
def check_writeable(target: str) -> None:
  if status.is_directory_locked():
    print(f'''
It looks like you have another lab running in this directory. netlab cannot
create {target} as that might impact your ability to shut down
the other lab.

Please use 'netlab status' to check the status of labs running on this machine, or
'netlab down' to shut down the other lab running in this directory.

If you are sure that no other lab is running in this directory, you can remove
the netlab.lock file manually and retry.
''')
    common.fatal('Cannot create configuration files in a locked directory')

class _TopologyOutput(Callback):
  def __init__(self, output: str, data: Box) -> None:
    self.settings = data

    output_parts = output.split('=')

    output_list = output_parts[0].split(':')
    self.output = output_list[0]
    self.format = output_list[1:]

    if len(output_parts) > 1:
      self.filenames = output_parts[1].split(',')

  @classmethod
  def load(self, output: str, data: Box) -> typing.Optional['_TopologyOutput']:
    module_name = __name__+"."+re.split(':|=',output)[0]
    obj = self.find_class(module_name)
    if obj:
      return obj(output,data)
    else:
      return None

  def get_template_path(self) -> str:
    fmtname = self.format[0] if self.format else "default"
    return 'templates/outputs/' + self.output + "/" + fmtname + ".j2"

  def get_output_name(self, fname: str, topology: Box) -> typing.Optional[str]:
    if fname:
      return fname

    if self.settings:
      return self.settings.filename

    return None

  def write(self, topology: Box) -> None:
    common.fatal('someone called the "write" method of TopologyOutput abstract class')



#
# VRF module
#
import typing, re
import netaddr
from box import Box

from . import _Module,_routing,_dataplane,get_effective_module_attribute
from .. import common
from .. import data
from ..data import get_from_box,global_vars
from ..data.validate import validate_attributes
from ..data.types import must_be_list,must_be_dict,must_be_id
from ..augment import devices,groups,links
from .. import addressing

def populate_vrf_static_ids(topology: Box) -> None:
  for k in ('id','rd'):
    _dataplane.create_id_set(f'vrf_{k}')
    _dataplane.extend_id_set(f'vrf_{k}',_dataplane.build_id_set(topology,'vrfs',k,'topology'))

  _dataplane.set_id_counter('vrf_id',1,4095)

  for n in topology.nodes.values():
    for k in ('id','rd'):
      _dataplane.extend_id_set(f'vrf_{k}',_dataplane.build_id_set(n,'vrfs',k,f'nodes.{n.name}'))

#
# Get a usable AS number. Try bgp.as then vrf.as from node and global settings
#
def get_rd_as_number(obj: Box, topology: Box) -> typing.Optional[typing.Any]:
  return \
    get_from_box(obj,'bgp.as') or \
    get_from_box(obj,'vrf.as') or \
    get_from_box(topology,'bgp.as') or \
    get_from_box(topology,'vrf.as')

#
# Parse rd/rt value -- check whether the RD/RT value is in N:N format
#

def parse_rdrt_value(value: str) -> typing.Optional[typing.List[typing.Union[int,str]]]:
  try:
    (asn,vid) = str(value).split(':')
  except Exception as ex:
    return None

  try:
    return [int(asn),int(vid)]
  except Exception as ex:
    try:
      netaddr.IPNetwork(asn)
      return [asn,int(vid)]
    except Exception as ex:
      return None

def get_next_vrf_id(asn: str) -> typing.Tuple[int,str]:
  rd_set = _dataplane.get_id_set('vrf_rd')
  id_set = _dataplane.get_id_set('vrf_rd')
  while True:
    vrf_id = _dataplane.get_next_id('vrf_id')
    if not f'{asn}:{vrf_id}' in rd_set:
      break

  rd = f'{asn}:{vrf_id}'
  rd_set.add(rd)
  id_set.add(vrf_id)
  return (vrf_id,rd)

#
# Normalize VRF IDs -- give a set of VRFs, change integer values of RDs into N:N strings
# Also checks for valid naming
#
def normalize_vrf_dict(obj: Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return

  asn = None
  obj_name = 'global VRFs' if obj is topology else obj.name

  if not isinstance(obj.vrfs,dict):
    common.error(f'VRF definition in {obj_name} is not a dictionary',common.IncorrectValue,'vrf')
    return

  for vname in list(obj.vrfs.keys()):
    must_be_id(parent=None,key=vname,path=f'NOATTR:VRF name {vname} in {obj_name}',module='vrf')

    if obj.vrfs[vname] is None:
      obj.vrfs[vname] = {}
    if not isinstance(obj.vrfs[vname],dict):
      common.error(f'VRF definition for {vname} in {obj_name} should be empty or a dictionary',
        common.IncorrectValue,
        'vrf')
      return

    vdata = obj.vrfs[vname]
    if 'rd' in vdata:
      if vdata.rd is None:      # RD set to None can be used to auto-generate RD while preventing RD inheritance
        continue                # ... skip the rest of the checks
      if isinstance(vdata.rd,int):
        asn = asn or get_rd_as_number(obj,topology)
        if not asn:
          common.error(f'VRF {vname} in {obj_name} uses integer RD value without a usable vrf.as or bgp.as AS number',
            common.MissingValue,
            'vrf')
          return
        vdata.rd = f'{asn}:{vdata.rd}'
      elif isinstance(vdata.rd,str):
        if parse_rdrt_value(vdata.rd) is None:
          common.error(f'RD value in VRF {vname} in {obj_name} ({vdata.rd}) is not in N:N format',
            common.IncorrectValue,
            'vrf')
      else:
        common.error(f'RD value in VRF {vname} in {obj_name} must be a string or an integer',
          common.IncorrectValue,
          'vrf')

def normalize_vrf_ids(topology: Box) -> None:
  normalize_vrf_dict(topology,topology)

  for n in topology.nodes.values():
    normalize_vrf_dict(n,topology)

def vrf_needs_id(vrf: Box) -> bool:
  if 'rd' in vrf and 'id' in vrf:
    return False
  return True

def set_vrf_auto_id(vrf: Box, value: typing.Tuple[int,str]) -> None:
  if not 'id' in vrf:
    vrf.id = value[0]

  if not 'rd' in vrf:
    vrf.rd = value[1]

#
# Get VRF RD value needed for import/export values. 
#
# WARNING: Global value takes precedence over node value because you might want to change per-node RD
# values for weird topologies like hub-and-spoke
#
def get_vrf_id(vname: str, obj: Box, topology: Box) -> typing.Optional[str]:
  obj_name = 'global VRFs' if obj is topology else obj.name
  vdata = get_from_box(topology,['vrfs',vname]) or get_from_box(obj,['vrfs',vname]) or None

  if vdata is None:
    common.error(
      f'Cannot get VRF ID for unknown VRF {vname} needed in {obj_name}',
      common.MissingValue,
      'vrf')
    return None

  if not isinstance(vdata,Box):
    common.fatal(f'Internal error: got a VRF definition that is not a dictionary')
    return None

  if not 'rd' in vdata:
    common.fatal(f'Internal error: VRF {vname} in {obj_name} should have a RD value by now')
    return None

  return vdata.rd

#
# Set RD values for all VRFs that have no RD attribute or RD value set to None (= auto-generate)
#
def set_vrf_ids(obj: Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return

  asn = None
  obj_name = 'global VRFs' if obj is topology else obj.name

  for vname,vdata in obj.vrfs.items():
    if vrf_needs_id(vdata):
      asn = asn or get_rd_as_number(obj,topology)
      if not asn:
        common.error('Need a usable vrf.as or bgp.as to create auto-generated VRF RD for {vname} in {obj_name}',
          common.MissingValue,
          'vrf')
        return
      set_vrf_auto_id(vdata,get_next_vrf_id(asn))

#
# Set import/export route targets
#
def set_import_export_rt(obj : Box, topology: Box) -> None:
  if not 'vrfs' in obj:
    return None

  obj_name = 'global VRFs' if obj is topology else obj.name
  obj_id   = 'vrfs' if obj is topology else f'nodes.{obj.name}.vrfs'
  asn      = None

  for vname,vdata in obj.vrfs.items():
    for rtname in ['import','export']:
      if not rtname in vdata:
        vdata[rtname] = [ vdata.rd ]
        continue

      must_be_list(vdata,rtname,f'{obj_id}.{vname}')

      rtlist = []     # The final parsed and looked-up list of RT values
      for rtvalue in vdata[rtname]:
        if isinstance(rtvalue,int):         # RT can be specified as an integer, in which case ASN is prepended to it
          asn = asn or get_rd_as_number(obj,topology)
          if not asn:
            common.error('VRF {vname} in {obj_id} uses integer {rtname} value without a usable vrf.as or bgp.as AS number',
              common.MissingValue,
              'vrf')
            continue
          rtvalue = f'{asn}:{rtvalue}'
        elif not isinstance(rtvalue,str):   # If RT is not an integer, it really should be a string
          common.error('{rtname} value {rtvalue} in VRF {vname} in {obj_id} should be a string or an integer',
            common.IncorrectValue,
            'vrf')
          continue
        else:
          if ':' in rtvalue:                # If there's a colon in RT value, then we're assuming N:N format
            if parse_rdrt_value(rtvalue) is None:
              common.error('{rtname} value {rtvalue} in VRF {vname} in {obj_id} is not in valid N:N format',
                common.IncorrectValue,
                'vrf')
              continue
          else:                             # Otherwise the RT value should refer to another VRF name
            rtvalue = get_vrf_id(rtvalue,obj,topology)
            if rtvalue is None:
              continue            # Error message generated in get_vrf_id

        rtlist.append(rtvalue)

      vdata[rtname] = rtlist

#
# VRF route leaking is usually implemented through BGP VPNv4 address families
# Check whether we have BGP AS configured on all nodes that use VRFs with route leaking
# (identified as import or export RT not equal to [ RD ])
#

def validate_vrf_route_leaking(node : Box) -> None:
  for vname,vdata in node.vrfs.items():
    simple_rt = [ vdata.rd ]
    leaked_routes = vdata['import'] and vdata['import'] != simple_rt
    leaked_routes = leaked_routes or (vdata['export'] and vdata['export'] != simple_rt)
    if leaked_routes:
      if not get_from_box(node,'bgp.as'):
        if get_from_box(node,'vrf.as'):
          node.bgp['as'] = node.vrf['as']
        else:
          common.error(
            f"VRF {vname} on {node.name} uses inter-VRF route leaking, but there's no BGP AS configured on the node",
            common.MissingValue,
            'vrf')

def vrf_loopbacks(node : Box, topology: Box) -> None:
  loopback_name = devices.get_device_attribute(node,'loopback_interface_name',topology.defaults) or \
                  devices.get_device_attribute(node,'features.vrf.loopback_interface_name',topology.defaults)

  if not loopback_name:                                                        # pragma: no cover -- hope we got device settings right ;)
    common.print_verbose(f'Device {node.device} used by {node.name} does not support VRF loopback interfaces - skipping assignment.')
    return

  node_vrf_loopback = get_effective_module_attribute(
                        path = 'vrf.loopback',
                        node = node,
                        topology = topology)
  for vrfname,v in node.vrfs.items():
    vrf_loopback = get_from_box(v,'loopback') or node_vrf_loopback          # Do we have VRF loopbacks enabled in the node or in the VRF?
    if not vrf_loopback:                                                    # ... nope, move on
      continue

    ifdata = Box({
      'virtual_interface': True,
      'type': "loopback",
      'name': f'VRF Loopback {vrfname}',
      'ifindex': node.interfaces[-1].ifindex + 1,
      'ifname': loopback_name.format(vrfidx=v.vrfidx,ifindex=v.vrfidx),     # Use VRF-specific and generic loopback index
      'neighbors': [],
      'vrf': vrfname,
    },default_box=True,box_dots=True)

    if isinstance(vrf_loopback,bool):
      vrfaddr = addressing.get(topology.pools, ['vrf_loopback'])
    else:
      vrfaddr = addressing.parse_prefix(vrf_loopback)

    if not vrfaddr:
      continue

    ospf_area = get_effective_module_attribute(
                  path = 'ospf.area',
                  link = v,
                  node = node,
                  topology = topology)

    if ospf_area:
      ifdata.ospf.area = ospf_area

    for af in vrfaddr:
      if af == 'ipv6':
        ifdata[af] = addressing.get_addr_mask(vrfaddr[af],1)
      else:
        ifdata[af] = str(vrfaddr[af])
      vrfaddr[af] = str(ifdata[af])                                         # Save string copy in vrfaddr, we need it later
      node.vrfs[vrfname].af[af] = True                                      # Enable the af if not already

    if not 'networks' in v:                                                 # List of networks to advertise in VRF BGP instance
      v.networks = []

    v.networks.append(vrfaddr)
    node.interfaces.append(ifdata)

    # add loopback addresses to the vrf data as well
    v.loopback_address = vrfaddr

  return

"""
create_vrf_links -- create VRF links based on VRF 'links' attribute

* Iterate over global VRFs
* If a VRF has 'links' attribute, verify that it's a list
* Iterate over the 'links' list
* Normalize every link in the list, add 'vrf: vname' attribute and append the link
  to global list of links
"""

def create_vrf_links(topology: Box) -> None:
  if not 'vrfs' in topology:                                                # No global VRFs, nothing to do
    return

  for vname,vdata in topology.vrfs.items():                                 # Iterate over global VRFs
    if not isinstance(vdata,Box):                                           # VRF not yet a dictionary?
      continue                                                              # ... no problem, skip it
    if not 'links' in vdata:                                                # No VRF links?
      continue                                                              # ... no problem, move on

    try:
      must_be_list(                                                         # Verify that the 'links' attribute is a list
        parent=vdata,
        key='links',
        path=f'vlans.{vname}',
        create_empty=False,
        module='vlans',
        abort=True)
    except:                                                                 # Error: not a list
      vdata.pop('links',None)                                               # ... remove the attribute
      continue                                                              # ... and move on

    for cnt,l in enumerate(vdata.links):                                    # So far so good, now iterate over the links
      link_data = links.adjust_link_object(                                 # Create link data from link definition
                    l=l,
                    linkname=f'vrfs.{vname}.links[{cnt+1}]',
                    nodes=topology.nodes)
      if link_data is None:
        continue
      link_data.vrf = vname                                                 # ... add VRF
      link_data.linkindex = links.get_next_linkindex(topology)              # ... add linkindex (we're late in the process)
      topology.links.append(link_data)                                      # ... and append new link to global link list

    vdata.pop('links')                                                      # Finally, clean up the VLAN definition

class VRF(_Module):

  def module_pre_default(self, topology: Box) -> None:
    for attr_set in ['global','node']:
      if not 'vrfs' in topology.defaults.attributes[attr_set]:
        topology.defaults.attributes[attr_set].append('vrfs')

  def module_pre_transform(self, topology: Box) -> None:
    if 'vrfs' in topology:
      try:
        must_be_dict(
          parent=topology,
          key='vrfs',
          path='topology',
          create_empty=False,
          abort=True,
          module='vrf')  # Check that we're dealing with a VRF dictionary and return otherwise
      except:
        return

    if 'groups' in topology:
      groups.export_group_node_data(topology,'vrfs','vrf',copy_keys=['rd','export','import'])

    if not 'vrfs' in topology:                          # No global VRFs, nothing to do
      return

    create_vrf_links(topology)                          # Create VRF links (and remove 'links' attribute)
    for vname in topology.vrfs.keys():
      must_be_dict(
        parent=topology,
        key=f'vrfs.{vname}',
        path='topology',
        create_empty=True,
        module='vrf')

      vdata = topology.vrfs[vname]
      validate_attributes(
        data=vdata,                                     # Validate global VRF data
        topology=topology,
        data_path=f'vrfs.{vname}',                      # Path to global VRF definition
        data_name=f'VRF',
        attr_list=['vrf','link'],                       # We're checking VLAN and link attributes
        modules=topology.get('module',[]),              # ... against global modules
        module_source='topology',
        module='vrf')                                   # Function is called from 'vrf' module

    common.exit_on_error()
    normalize_vrf_ids(topology)
    populate_vrf_static_ids(topology)
    set_vrf_ids(topology,topology)
    set_import_export_rt(topology,topology)

  def node_pre_transform(self, node: Box, topology: Box) -> None:
    # Check if any global vrfs need to be pulled in due to being referenced by a vlan
    vlan_vrfs = [ vdata.vrf for vname,vdata in node.get('vlans',{}).items() if 'vrf' in vdata ]
    if not 'vrfs' in node:
      if not vlan_vrfs:  # No local vrfs and no vlan references -> exit
        return
      node.vrfs = {}     # Prepare to pull in global vrfs

    if not must_be_dict(
        parent=node,
        key='vrfs',
        path=f'nodes.{node.name}',
        create_empty=False,
        module='vrf'):                                # Check that we're dealing with a VRF dictionary and return if there's none
      return

    for vname in set(list(node.vrfs.keys()) + vlan_vrfs):  # Filter out duplicates
      if node.vrfs[vname] is None:
        node.vrfs[vname] = {}

      validate_attributes(
        data=node.vrfs[vname],                        # Validate node VRF data
        topology=topology,
        data_path=f'nodes.{node.name}.vrfs.{vname}',  # Path to node VRF definition
        data_name=f'VRF',
        attr_list=['vrf','link'],                     # We're checking VLAN and link attributes
        modules=node.get('module',[]),                # ... against node modules
        module_source=f'nodes.{node.name}',
        module='vrf')                                 # Function is called from 'vrf' module

      if 'vrfs' in topology and vname in topology.vrfs:
        node.vrfs[vname] = topology.vrfs[vname] + node.vrfs[vname]

    set_vrf_ids(node,topology)
    set_import_export_rt(node,topology)

  def link_pre_transform(self, link: Box, topology: Box) -> None:
    pass

  #
  # The post-link-transform hook must normalize VRF data and pull global VRF data into
  # nodes so we can use the node VRF data in copy_node_data_into_interfaces module function
  #
  # We have to iterate over all VRF interfaces, validate VRF names (which was previously done
  # in post-transform hook), and populate node VRF data (moved from post-transform hook)
  #
  def node_post_link_transform(self, node: Box, topology: Box) -> None:
    for ifdata in node.interfaces:
      if not 'vrf' in ifdata:                                           # Check only VRF interfaces
        continue

      vrf_data_path = f'vrfs.{ifdata.vrf}'
      if not get_from_box(topology,vrf_data_path) and not get_from_box(node,vrf_data_path):
        common.error(
          f'VRF {ifdata.vrf} used on an interface in {node.name} is not defined in the node or globally',
          common.MissingValue,
          'vrf')
        continue

      if not ifdata.vrf in node.vrfs:                                   # Local VRF not present, mark as required
        node.vrfs[ifdata.vrf] = {}

    if not 'vrfs' in node:
      return

    for vname in node.vrfs.keys():
      if vname in topology.get('vrfs',{}):                              # Carefully check for global VRF
        node.vrfs[vname] = topology.vrfs[vname] + node.vrfs[vname]      # ... and do the data merge

  def node_post_transform(self, node: Box, topology: Box) -> None:
    vrf_count = 0

    for ifdata in node.interfaces:
      if 'vrf' in ifdata:
        vrf_count = vrf_count + 1
        if not node.vrfs[ifdata.vrf].rd:
          common.error(
            f'VRF {ifdata.vrf} used on an interface in {node.name} does not have a usable RD',
            common.MissingValue,
            'vrf')
          continue

        for af in ['v4','v6']:
          if f'ip{af}' in ifdata:
            node.af[f'vpn{af}'] = True
            node.vrfs[ifdata.vrf].af[f'ip{af}'] = True

    if common.debug_active('vrf'):
      print( f"vrf node_post_transform on {node.name}: counted {vrf_count} VRFs on interfaces" )
    features = devices.get_device_features(node,topology.defaults)
    if not vrf_count and ('vrf' not in features or not features.vrf.keep_module): # Remove VRF module from the node if the node has no VRFs, unless flag set
      node.module = [ m for m in node.module if m != 'vrf' ]
      node.pop('vrfs',None)
    else:
      node.vrfs = node.vrfs or {}     # ... otherwise make sure the 'vrfs' dictionary is not empty
      vrfidx = 100

      # Check that all VRFs have a well-defined data structure (should be at this point, unless someone used groups.node_data)
      for k,v in node.vrfs.items():
        if v is None or not 'id' in v:
          common.error(
            f"Found invalid VRF {k} on node {node.name}. Did you mention it only in groups.node_data? You can't do that.",
            common.IncorrectValue,
            'vrf')
          common.exit_on_error()

      # We need unique VRF index to create OSPF processes, assign in order sorted by VRF ID "for consistency"
      for v in sorted(node.vrfs.values(),key=lambda v: v.id):
        v.vrfidx = vrfidx
        vrfidx = vrfidx + 1

      validate_vrf_route_leaking(node)

      # Set additional loopbacks (one for each defined VRF)
      vrf_loopbacks(node, topology)

    # Finally, set BGP router ID if we set BGP AS number
    #
    if get_from_box(node,'bgp.as') and not get_from_box(node,'bgp.router_id'):
      _routing.router_id(node,'bgp',topology.pools)



#
# Cisco ASA quirks
#
from box import Box

from . import _Quirks
from .. import common
from ..data import get_from_box

def check_isis_p2p_interfaces(node: Box, topology: Box) -> None:
  for intf in node.interfaces:
    if not 'isis' in intf:
      continue
    for neighbor in intf.neighbors:
      remote_node = neighbor.node
      remote_interfaces = topology.nodes[remote_node].interfaces
      for rintf in remote_interfaces:
        if rintf.ifname == neighbor.ifname:
          if get_from_box(rintf, 'isis.network_type') == "point-to-point":
            common.error(
                f'Cisco ASA does not support P2P IS-IS links.'
                f'Problematic Interface: {remote_node} {neighbor.ifname}',
                common.IncorrectType,
                'quirks',
            )

class ASA(_Quirks):

  @classmethod
  def device_quirks(self, node: Box, topology: Box) -> None:
    mods = node.get('module', [])
    if 'isis' in mods:
      check_isis_p2p_interfaces(node, topology)



'''
plugin - implement custom topology transformation plugins
'''

import os
import sys
import typing
import importlib
import importlib.util

from box import Box
from .. import common
from .. import data

def load_plugin_from_path(path: str, plugin: str) -> typing.Optional[object]:
  module_path = path+'/'+plugin
  if not os.path.exists(module_path):
    if os.path.exists(module_path+'.py'):
      module_path = module_path+'.py'
    else:
      return None

  module: typing.Optional[object] = None
  config_name = None

  if os.path.isdir(module_path):
    module_path = module_path + '/plugin.py'
    config_name = plugin
    if not os.path.exists(module_path):
      return None

  module_name = plugin.replace('.py','')
  module_name = f'netlab.plugin.{module_name}'

  try:
    modspec  = importlib.util.spec_from_file_location(module_name,module_path)
    assert(isinstance(modspec,importlib.machinery.ModuleSpec))
    pymodule = importlib.util.module_from_spec(modspec)
    sys.modules[module_name] = pymodule
    assert(isinstance(modspec.loader,importlib.abc.Loader))
    modspec.loader.exec_module(pymodule)
  except:
    print(f"Cannot load plugin {module_name} from {module_path}\n{str(sys.exc_info()[1])}")
    common.fatal('Aborting the transformation process','plugin')

  if config_name:
    setattr(pymodule,'config_name',config_name)

  if common.VERBOSE:
    print(f"loaded plugin {module_name}")
  return pymodule

def init(topology: Box) -> None:
  data.types.must_be_list(parent=topology,key='defaults.plugin',path='',create_empty=True)    # defaults.plugin must be a list (if present)
  if topology.defaults.plugin:                                                                # If we have default plugins...
    data.types.must_be_list(parent=topology,key='plugin',path='',create_empty=True)           # ... make sure the plugin attribute is a list
    topology.plugin.extend(topology.defaults.plugin)                                          # ... and extend it with default plugins

  if not 'plugin' in topology:
    return

  topology.Plugin = []
  for pname in topology.plugin:
    plugin = None
    search_path = ('.',common.netsim_package_path+'/extra')
    for path in search_path:
      if not plugin:
        plugin = load_plugin_from_path(path,pname)
    if plugin:
      topology.Plugin.append(plugin)
    else:
      common.error(f"Cannot find plugin {pname} in {search_path}",common.IncorrectValue,'plugin')

  if common.debug_active('plugin'):
    print(f'plug INIT: {topology.Plugin}')

def execute(action: str, topology: Box) -> None:
  if not 'Plugin' in topology:
    return

  for plugin in topology.Plugin:
    if hasattr(plugin,action):
      func = getattr(plugin,action)
      if common.debug_active('plugin'):
        print(f'plug INIT: {topology.Plugin}')
      func(topology)



# Import the whole augmentation tree

from . import nodes
from . import groups
from . import links
from . import devices
from . import plugin
from . import topology
from . import main
from . import config



#
# Vagrant/libvirt provider module
#
import typing
from box import Box

from . import _TopologyOutput,check_writeable
from .. import providers
from .. import common
from ..augment import nodes

def get_provider_module(topology: Box, provider: str) -> providers._Provider:
  return providers._Provider.load(provider,topology.defaults.providers[provider])

def write_provider_file(topology: Box, provider: str, filename: typing.Optional[str]) -> None:
  p_module = get_provider_module(topology,provider)
  p_module.create(topology,filename)

class ProviderConfiguration(_TopologyOutput):

  def write(self, topology: Box) -> None:
    check_writeable('provider configuration')
    filename = None
    if hasattr(self,'filenames'):
      filename = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,'provider')

    if self.format:
      common.error('Specified output format(s) %s ignored' % self.format,common.IncorrectValue,'provider')

    # Creates a "ghost clean" topology after transformation
    # (AKA, remove unmanaged devices)
    topology = nodes.ghost_buster(topology)
    p_module = get_provider_module(topology,topology.provider)
    providers.mark_providers(topology)
    p_module.call('pre_output_transform',topology)
    write_provider_file(providers.select_topology(topology,topology.provider),topology.provider,filename)

    for subprovider in topology[topology.provider].providers.keys():  # Iterate over subproviders
      print(f"Creating configuration file for secondary provider {subprovider}")
      write_provider_file(
        topology=providers.select_topology(topology,subprovider),
        provider=subprovider,
        filename=topology.defaults.providers[topology.provider][subprovider].filename)


#
# Common routines for create-topology script
#
import textwrap
from box import Box

def extra_data_printout(s : str, width: int = 70) -> str:
  lines = []
  for line in s.split('\n'):
    lines.append(textwrap.TextWrapper(
      initial_indent="... ",
      subsequent_indent="      ",
      width=width).fill(line))

  return "\n".join(lines)

def format_structured_dict(d: Box, prefix: str = '') -> str:
  lines = []
  for k,v in d.items():
    if v and (isinstance(v,dict) or isinstance(v,list)):
      lines.append(f'{prefix}{k}:')
      lines.append(f'{prefix}  {v}')
    else:
      lines.append(f'{prefix}{k}: {v}')

  return "\n".join(lines)

def print_structured_dict(d: Box, prefix: str = '') -> None:
  print(format_structured_dict(d,prefix))

#
# eval_format: emulate f'strings' evaluated on a data structure
#
def eval_format(fmt: str, data: dict) -> str:
  return str(eval(f"f'{fmt}'",dict(data)))                            # An awful hack to use f-string specified in a string variable

"""
confirm: print the prompt and wait for a yes/no answer
"""
def confirm(prompt: str) -> bool:
  prompt = f'{prompt} [y/n]'

  while True:
    answer = input(prompt).lower()
    if answer in ['y','yes']:
      return True
    if answer in ['n','no']:
      return False



#
# Utility routines to augment configuration files
#

import typing
from box import Box

from .. import common

"""
Augment module attributes

* Copy data types from global to nodes (unless no_propagate is set)
* Copy data types from links to interfaces
* Copy data types from node to interfaces if an attribute is in the node_copy list
"""

def copy_datatypes(
      parent: typing.Optional[Box],                                   # Parent attributes (example: global)
      child: typing.Optional[Box],                                    # Child attributes (example: node)
      copy_list: list = [],                                           # Only copy these attributes (example: node_copy)
      skip_list: list = []) -> None:                                  # Skip these attributes (example: module.no_propagate)

  if not isinstance(child,Box) or not isinstance(parent,Box):         # Datatype propagation works only for box-to-box case
    return

  for k in child.keys():
    if copy_list and not k in copy_list:                              # If we have a copy list and the attribute is not in it, skip it
      continue

    if skip_list and k in skip_list:                                  # If we have a skip list and the attribute is in it, skip it
      continue

    if child[k] is None and k in parent:
      child[k] = parent[k]

def adjust_attributes(
      attr: Box,
      global_no_propagate: list = [],
      link_no_propagate: list = [],
      node_copy: list = []) -> None:

  copy_datatypes(                                                     # Copy global datatypes to node datatyps
    parent=attr.get('global',None),
    child=attr.get('node',None),
    skip_list=global_no_propagate)
  copy_datatypes(                                                     # Copy link datatypes to interface datatypes
    parent=attr.get('link',None),
    child=attr.get('interface',None),
    skip_list=link_no_propagate)
  if node_copy:                                                       # Copy node datatypes to interface datatypes
    copy_datatypes(                                                   # ... if the node_copy list is defined
      parent=attr.get('node',None),
      child=attr.get('interface',None),
      copy_list=node_copy)


def attributes(topology: Box) -> None:
  defaults=topology.defaults

  for modname,moddata in defaults.items():                            # Iterate over top-level defaults
    if not isinstance(moddata,Box) or not 'supported_on' in moddata:  # Skip everything that is not a module
      continue
    adjust_attributes(
      attr=moddata.attributes,
      global_no_propagate=moddata.get('no_propagate',[]),
      link_no_propagate=moddata.attributes.get('link_no_propagate',[]),
      node_copy=moddata.attributes.get('node_copy',[]))



#
# netlab down command
#
# * Transform lab topology or read transformed lab topology from snapshot file
# * Stop the lab, including provider-specific pre- and post-stop hooks
# * Clean up the working directory (optional)
#
import argparse
import typing
import textwrap
import os
import sys
from box import Box

from . import topology_parse_args, load_snapshot_or_topology, external_commands,fs_cleanup
from . import lab_status_change, lab_status_update,get_lab_id
from .. import read_topology,augment,common
from .. import providers
from ..utils import status,strings
from .up import provider_probes
#
# CLI parser for 'netlab down' command
#
def down_parse(args: typing.List[str]) -> argparse.Namespace:
  parser = argparse.ArgumentParser(
    parents=[ topology_parse_args() ],
    prog="netlab down",
    description='Destroy the virtual lab')

  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default = 0,
    help='Verbose logging (where applicable)')
  parser.add_argument(
    '--cleanup',
    dest='cleanup',
    action='store_true',
    help='Remove all configuration files created by netlab create')
  parser.add_argument(
    '--snapshot',
    dest='snapshot',
    action='store',
    nargs='?',
    const='netlab.snapshot.yml',
    help='Transformed topology snapshot file')
  parser.add_argument(
    dest='topology', action='store', nargs='?',
    help='Topology file (default: topology.yml)')

  return parser.parse_args(args)

def down_cleanup(topology: Box, verbose: bool = False) -> None:
  p_provider = topology.provider
  cleanup_list = topology.defaults.providers[p_provider].cleanup or []

  for s_provider in topology[p_provider].providers:
    cleanup_list.extend(topology.defaults.providers[s_provider].cleanup or [])
    s_filename = topology.defaults.providers[p_provider][s_provider].filename
    if s_filename:
      cleanup_list.append(s_filename)

  cleanup_list.extend(topology.defaults.automation.ansible.cleanup)
  cleanup_list.append('netlab.snapshot.yml')
  fs_cleanup(cleanup_list,verbose)

def stop_provider_lab(topology: Box, pname: str, sname: typing.Optional[str] = None) -> None:
  p_name = sname or pname
  p_topology = providers.select_topology(topology,p_name)
  p_module   = providers._Provider.load(p_name,topology.defaults.providers[p_name])

  exec_command = None
  if sname is not None:
    exec_command = topology.defaults.providers[pname][sname].stop

  p_module.call('pre_stop_lab',p_topology)
  external_commands.stop_lab(topology.defaults,p_name,2,"netlab down",exec_command)
  p_module.call('post_stop_lab',p_topology)

'''
lab_dir_mismatch -- check if the lab instance is running in the current directory
'''
def lab_dir_mismatch(topology: Box) -> bool:
  lab_id = get_lab_id(topology)
  lab_status = status.read_status(topology)       # Find current running instance(s)
  if not lab_id in lab_status:                    # This could be a lab instance from pre-status days
    return False                                  # ... in which case we can shut it down
  if lab_id in lab_status and lab_status[lab_id].dir == os.getcwd():
    return False                                 # Lab instance is known  and this is the correct directory        

  print(f'''
According to the netlab status file, the lab instance '{lab_id}' is running
in directory {lab_status[lab_id].dir}.

You could proceed if you want to clean up the netlab artifacts from this
directory, but you might impact the running lab instance.
''')
  if not strings.confirm('Do you want to proceed?'):
    common.fatal('aborting lab shutdown request')

  return True

'''
Remove the lab instance/directory from the status file
'''
def remove_lab_status(topology: Box) -> None:
  lab_id = get_lab_id(topology)

  status.change_status(
    topology,
    callback = lambda s,t: s.pop(lab_id,None))

def run(cli_args: typing.List[str]) -> None:
  args = down_parse(cli_args)
  topology = load_snapshot_or_topology(args)

  if args.topology:
    print(f"Reading lab topology from {args.topology}")
  else:
    if not os.path.isfile(args.snapshot):
      print(f"The topology snapshot file {args.snapshot} does not exist.\n"+
            "Looks like no lab was started from this directory")
      sys.exit(1)

    print(f"Reading transformed lab topology from snapshot file {args.snapshot}")

  if topology is None:
    common.fatal('... could not read the lab topology, aborting')
    return

  mismatch = lab_dir_mismatch(topology)

  lab_status_change(topology,f'lab shutdown requested{" in conflicting directory" if mismatch else ""}')
  provider_probes(topology)

  p_provider = topology.provider
  p_module = providers._Provider.load(p_provider,topology.defaults.providers[p_provider])
  providers.mark_providers(topology)
  p_module.call('pre_output_transform',topology)

  for s_provider in topology[p_provider].providers:
    lab_status_change(topology,f'stopping {s_provider} provider')
    stop_provider_lab(topology,p_provider,s_provider)
    print()

  stop_provider_lab(topology,p_provider)

  if args.cleanup:
    external_commands.print_step(3,"Cleanup configuration files",spacing = True)
    down_cleanup(topology,True)

  if not mismatch:
    remove_lab_status(topology)
  status.unlock_directory()



#
# Vagrant/libvirt provider module
#

import subprocess
import re
import sys
import os
import typing
from box import Box
import pathlib
import tempfile
import netaddr

from .. import common
from ..data import get_from_box,types
from . import _Provider
from ..augment.links import get_link_by_index

LIBVIRT_MANAGEMENT_NETWORK_NAME = "vagrant-libvirt"
LIBVIRT_MANAGEMENT_BRIDGE_NAME  = "libvirt-mgmt"
LIBVIRT_MANAGEMENT_NETWORK_FILE = "templates/provider/libvirt/vagrant-libvirt.xml"
LIBVIRT_MANAGEMENT_SUBNET       = "192.168.121.0/24"

"""
Replace management IP subnet in vagrant-libvirt XML template:

* Replace subnet (.1 address) and netmask
* Replace start (.2) and end (start -1) of dynamic DHCP range
* Replace IP addresses in static DHCP bindings (from start until the next address is no longer found)

Replacements have to match single quotes (in XML) to ensure we don't replace partial IP addresses
"""

def replace_xml_mgmt_subnet(xml: str, mgmt: Box, m_subnet: str) -> str:
  o_net = netaddr.IPNetwork(m_subnet)
  d_net = netaddr.IPNetwork(mgmt.ipv4)

  xml = xml.replace(f"'{o_net.netmask}'",f"'{d_net.netmask}'")
  for offset in [1,2]:
    xml = xml.replace(f"'{o_net[offset]}'",f"'{d_net[offset]}'")

  o_start = 100
  d_start = mgmt.start
  mac_cnt = 0

  xml = xml.replace(f"'{o_net[o_start - 1]}'",f"'{d_net[d_start - 1]}'")
  while True:                                                         # Replace predefined static DHCP bindings
    o_start += 1
    d_start += 1
    mac_cnt += 1
    o_addr = str(o_net[o_start])

    if not o_addr in xml:
      break

    xml = xml.replace(f"'{o_addr}'",f"'{d_net[d_start]}'")

  eui = netaddr.EUI(mgmt.mac)
  while d_start < d_net.size - 2:
    eui[5] = mac_cnt
    xstring = f"<host mac='{str(eui).replace('-',':')}' ip='{d_net[d_start]}'/>\n<!--more-->"
    xml = xml.replace("<!--more-->",xstring)
    d_start += 1
    mac_cnt += 1

  return xml

"""
Create a virsh net-define XML file from vagrant-libvirt XML template:

* Replace network and bridge name if needed
* Replace IP subnet/mask and DHCP bindings
* Create a temporary file with modified XML definitions
* Return the name of the temporary file
"""

def get_libvirt_mgmt_template() -> str:
  return str(pathlib.Path(__file__).parent.parent.joinpath(LIBVIRT_MANAGEMENT_NETWORK_FILE).resolve())

def create_network_template(topology: Box) -> str:
  net_template_xml = get_libvirt_mgmt_template()
  mgmt = topology.addressing.mgmt
  try:
    with open(net_template_xml) as xfile:
      xml = xfile.read()
  except Exception as ex:
    common.fatal(f'Cannot open/read XML definition of vagrant-libvirt network {str(sys.exc_info()[1])}')

  if mgmt._network:
    xml = xml.replace(LIBVIRT_MANAGEMENT_NETWORK_NAME,mgmt._network)

  if mgmt._bridge:
    xml = xml.replace(LIBVIRT_MANAGEMENT_BRIDGE_NAME,mgmt._bridge)

  xml = replace_xml_mgmt_subnet(xml,mgmt,LIBVIRT_MANAGEMENT_SUBNET)

  with tempfile.NamedTemporaryFile(mode='w',delete=False) as tfile:
    tfile.write(xml)
    tfile.close()
    return tfile.name

def create_vagrant_network(topology: typing.Optional[Box] = None) -> None:
  mgmt_net = topology.addressing.mgmt._network if topology is not None else ''
  mgmt_net = mgmt_net or LIBVIRT_MANAGEMENT_NETWORK_NAME
  try:
    subprocess.run(['virsh','net-destroy',mgmt_net],capture_output=True,text=True,check=False)    # Remove management network
    subprocess.run(['virsh','net-undefine',mgmt_net],capture_output=True,text=True,check=False)   # ... if it exists
    common.print_verbose(f'creating libvirt management network {mgmt_net}')

    if topology is None:
      net_template = get_libvirt_mgmt_template()                    # When called without topology data use the default template
    else:
      net_template = create_network_template(topology)              # Otherwise create a temporary XML file
    result2 = subprocess.run(['virsh','net-define',net_template],capture_output=True,check=True,text=True)
    if not topology is None:                                        # Remove the temporary XML file if needed
      os.remove(net_template)

  except subprocess.CalledProcessError as e:
    common.fatal(f'Exception in net handling for libvirt network {mgmt_net}: {e.returncode}\n{e.stderr}')
  return

def get_linux_bridge_name(virsh_bridge: str) -> typing.Optional[str]:
  try:
    result = subprocess.run(['virsh','net-info',virsh_bridge],capture_output=True,check=True,text=True)
  except:
    common.error('Cannot run net-info for libvirt network %s' % virsh_bridge, module='libvirt')
    return None

  match = None
  if result and result.returncode == 0:
    match = re.search("Bridge:\\s+(.*)$",result.stdout,flags=re.MULTILINE)

  if match:
    return match.group(1)
  else:
    common.error(f'Cannot get Linux bridge name for libvirt network {virsh_bridge}', module='libvirt')

  return None

"""
Create batches of 'vagrant up' command to deal with very large topologies

* Split node names into libvirt.batch_size - sized batches
* Change libvirt.start command into a list of commands
"""
def create_vagrant_batches(topology: Box) -> None:
  libvirt_defaults = topology.defaults.providers.libvirt
  if not libvirt_defaults.batch_size:
    return

  types.must_be_int(libvirt_defaults,'batch_size','defaults.providers.libvirt',module='libvirt',min_value=1,max_value=50)
  types.must_be_int(libvirt_defaults,'batch_interval','defaults.providers.libvirt',module='libvirt',min_value=1,max_value=1000)
  common.exit_on_error()

  batch_size = libvirt_defaults.batch_size
  start_cmd  = libvirt_defaults.start
  libvirt_defaults.start = []
  node_list = list(topology.nodes.keys())

  while True:
    libvirt_defaults.start.append(start_cmd + " " + " ".join(node_list[:batch_size]))     # Add up to batch_size nodes to the start command
    if len(node_list) <= batch_size:
      break
    node_list = node_list[batch_size:]
    if libvirt_defaults.batch_interval:
      libvirt_defaults.start.append(f'sleep {libvirt_defaults.batch_interval}')

class Libvirt(_Provider):

  """
  post_transform hook: mark multi-provider links as LAN links
  """
  def pre_transform(self, topology: Box) -> None:
    _Provider.pre_transform(self,topology)
    if not 'links' in topology:
      return

    for l in topology.links:
      if get_from_box(l,'libvirt.provider'):
        l.type = 'lan'
        if not 'bridge' in l:
          l.bridge = "%s_%d" % (topology.name[0:10],l.linkindex)

  def transform_node_images(self, topology: Box) -> None:
    self.node_image_version(topology)

  def pre_output_transform(self, topology: Box) -> None:
    _Provider.pre_output_transform(self,topology)
    for link in topology.links:                                     # Adjust links to deal with subprovider gotchas
      if link.type != 'lan':                                        # Multi-provider links are always LAN links
        continue

      if len(link.provider) <= 1:                                   # Skip single-provider links
        continue

      if 'clab' in link.provider:                                   # Find links with clab subprovider
        link.node_count = 999                                       # ... and fake link count to force clab to use a bridge
        if 'libvirt' in link.provider:                              # If the link uses libvirt provider
          link.clab.external_bridge = True                          # ... then the Linux bridge will be create by vagrant-libvirt

    for node in topology.nodes.values():                            # Now find P2P tunnel links and create interface data needed for Vagrantfile
      for intf in node.interfaces:
        if not intf.get('linkindex',None):                          # Cannot get interface index, skip it
          continue
        if intf.get('virtual_interface',None):                      # Virtual interface, skip it
          continue

        link = get_link_by_index(topology,intf.linkindex)           # Get the link object based on intf linkindex
        if link is None:                                            # Weird, cannot find the link, skip it
          continue

        if not 'libvirt' in link.provider:                          # Not a libvirt link? skip it
          continue

        if len(link.provider) > 1:                                  # multi-provider link. Skip it.
          continue

        if len(link.interfaces) == 2:
          intf.libvirt.type = "tunnel"                              # ... found a true libvirt-only P2P link, set type to tunnel

        if intf.libvirt.type != 'tunnel':                           # The current link is not a tunnel link, move on
          continue

        link.pop("bridge",None)                                     # And now the real work starts. Pop the bridge attribute first

        remote_if_list = [ rif for rif in link.interfaces if rif.node != node.name or rif.ifindex != intf.ifindex ]
        if len(remote_if_list) != 1:                                # There should be only one remote interface attached to this link
          common.fatal(
            f'Cannot find remote interface for P2P link\n... node {node.name}\n... intf {intf}\n... link {link}\n... iflist {remote_if_list}')
          return

        remote_if = remote_if_list[0]                               # Get remote interface
        intf.remote_ifindex = remote_if.ifindex                     # ... and copy its ifindex
        intf.remote_id = topology.nodes[remote_if.node].id          # ... and node ID
        if not intf.remote_id:
          common.fatal(
            f'Cannot find remote node ID on a P2P link\n... node {node.name}\n... intf {intf}\n... link {link}')
          return

  def pre_start_lab(self, topology: Box) -> None:
    common.print_verbose('pre-start hook for libvirt')
    # Starting from vagrant-libvirt 0.7.0, the destroy actions deletes all the networking
    #  including the "vagrant-libvirt" management network.
    #  Let's re-create it if missing!
    create_vagrant_network(topology)
    create_vagrant_batches(topology)

  def post_start_lab(self, topology: Box) -> None:
    common.print_verbose('libvirt lab has started, fixing Linux bridges')
    mgmt_bridge = get_linux_bridge_name(topology.addressing.mgmt._network or LIBVIRT_MANAGEMENT_NETWORK_NAME)
    if mgmt_bridge:
      topology.addressing.mgmt._bridge = mgmt_bridge

    for l in topology.links:
      brname = l.get('bridge',None)
      if not brname:                                                # Link not using a Linux bridge
        continue
      if not 'libvirt' in l.provider:                               # Not a libvirt link, skip it
        continue

      if common.debug_active('libvirt'):
        print('libvirt post_start_lab: fixing Linux bridge for link {l}')

      linux_bridge = get_linux_bridge_name(brname)
      if linux_bridge is None:
        continue

      l.bridge = linux_bridge
      common.print_verbose(f"... network {brname} maps into {linux_bridge}")
      try:
        subprocess.run(['sudo','sh','-c',f'echo 0x4000 >/sys/class/net/{linux_bridge}/bridge/group_fwd_mask'],check=True)
      except:
        common.error(f"Cannot set forwarding mask on Linux bridge {linux_bridge}")
        continue
      common.print_verbose(f"... setting LLDP enabled flag on {linux_bridge}")



import typing, netaddr
from box import Box

from netsim import api

"""
Enable 'proxy_arp' attribute per VRF
"""
def init(topology: Box) -> None:
  topology.defaults.attributes.vrf.proxy_arp = None

"""
Add 'proxy_arp' flag and trigger custom config file for any interfaces that are
part of a VRF with proxy ARP enabled
"""
def post_transform(topology: Box) -> None:

  if not topology.get('vrfs',[]):
    return

  target_vrfs = [ v for v,d in topology.vrfs.items() if 'proxy_arp' in d ]

  config_name = api.get_config_name(globals())

  # Iterate over node[x].interfaces that are part of a VRF with proxy_arp enabled
  for n, ndata in topology.nodes.items():
    for i in ndata.interfaces:
      if i.type in ['lan','p2p','svi'] and 'vrf' in i and i.vrf in target_vrfs \
        and ('ipv4' in i or 'ipv6' in i):
          i.proxy_arp = True
          api.node_config(ndata,config_name)



#
# Build full-blown topology data structures (nodes, links, global parameter) from high-level topology
#

import sys

from box import Box

from .. import common
from .. import addressing
from .. import augment
from .. import providers
from .. import modules
from .. import devices as quirks
from ..data import global_vars

def transform_setup(topology: Box) -> None:
  global_vars.init(topology)
  augment.config.attributes(topology)
  augment.topology.check_required_elements(topology)
  topology.nodes = augment.nodes.create_node_dict(topology.nodes)
  if 'links' in topology:
    augment.links.links_init(topology)

  augment.devices.augment_device_settings(topology)
  augment.plugin.init(topology)                                         # Initialize plugins very early on in case they modify extra attributes
  augment.plugin.execute('init',topology)
  common.exit_on_error()

  augment.topology.extend_attribute_list(topology.defaults)             # Attributes have to be extended before group init
  augment.topology.extend_module_attribute_list(topology)               # ... or we won't recognize node attributes in groups
  augment.groups.init_groups(topology)
  common.exit_on_error()

  augment.topology.adjust_global_parameters(topology)
  common.exit_on_error()

  augment.nodes.augment_node_provider_data(topology)
  augment.nodes.augment_node_system_data(topology)
  common.exit_on_error()

  modules.pre_default(topology)
  common.exit_on_error()

  augment.topology.check_global_elements(topology)
  augment.nodes.validate(topology)
  common.exit_on_error()

def transform_data(topology: Box) -> None:
  addressing.setup(topology)
  common.exit_on_error()
  augment.plugin.execute('pre_transform',topology)
  modules.pre_transform(topology)
  providers.execute("pre_transform",topology)

  augment.plugin.execute('pre_node_transform',topology)
  modules.pre_node_transform(topology)
  augment.nodes.transform(topology,topology.defaults,topology.pools)
  common.exit_on_error()
  augment.plugin.execute('post_node_transform',topology)
  modules.post_node_transform(topology)

  if 'links' in topology:
    augment.links.validate(topology)
    common.exit_on_error()
    augment.plugin.execute('pre_link_transform',topology)
    modules.pre_link_transform(topology)
    common.exit_on_error()
    augment.links.transform(topology.links,topology.defaults,topology.nodes,topology.pools)
    common.exit_on_error()
    augment.plugin.execute('post_link_transform',topology)

  modules.post_link_transform(topology)

def post_transform(topology: Box) -> None:
  modules.post_transform(topology)
  augment.plugin.execute('post_transform',topology)
  augment.groups.node_config_templates(topology)
  providers.execute("post_transform",topology)
  common.exit_on_error()

  quirks.process_quirks(topology)
  common.exit_on_error()
  
  augment.links.cleanup(topology)
  for remove_attr in ['Plugin','pools','_Providers']:
    topology.pop(remove_attr,None)

def transform(topology: Box) -> None:
  transform_setup(topology)
  transform_data(topology)
  post_transform(topology)



import typing, netaddr
from box import Box
from netsim import common
from netsim import api
from netsim import data

"""
Adds a custom bgp.{allowas_in,as_override,default_originate} as link->node (interface) attribute
"""
def init(topology: Box) -> None:
    attr = topology.defaults.bgp.attributes
    attr.interface.allowas_in = { 'type': 'int', 'true_value': 1, 'min_value': 1, 'max_value': 10 }
    attr.interface.as_override = 'bool'
    attr.interface.default_originate = 'bool'
    attr.link.password = 'str'

def pre_link_transform(topology: Box) -> None:
    # Error if BGP module is not loaded
    if not 'bgp' in topology.module:
        common.error(
            f'BGP Module is not loaded.',
            common.IncorrectValue,
            'ebgp_utils')

def post_transform(topology: Box) -> None:
    config_name = api.get_config_name(globals())
    # Iterate over node[x].interfaces
    for n, ndata in topology.nodes.items():
        for intf in ndata.interfaces:
            #print("LINK START *******************")
            #print(intf)
            #print("LINK END   *******************")
            # Handle as_override
            as_override = data.get_from_box(intf, 'bgp.as_override')
            if as_override:
                # Report the parameter on the BGP session, based on ifindex
                for neigh in ndata.get('bgp', {}).get('neighbors', []):
                    if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                        neigh.as_override = as_override
                # Report the parameter on the BGP session (for VRF), based on ifindex
                if 'vrf' in intf:
                    for neigh in ndata.vrfs[intf.vrf].get('bgp', {}).get('neighbors', []):
                        if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                            neigh.as_override = as_override
                api.node_config(ndata,config_name)
            # Handle allowas_in
            allowas_in = data.get_from_box(intf, 'bgp.allowas_in')
            if allowas_in:
                # Report the parameter on the BGP session, based on ifindex
                for neigh in ndata.get('bgp', {}).get('neighbors', []):
                    if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                        neigh.allowas_in = int(allowas_in)
                # Report the parameter on the BGP session (for VRF), based on ifindex
                if 'vrf' in intf:
                    for neigh in ndata.vrfs[intf.vrf].get('bgp', {}).get('neighbors', []):
                        if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                            neigh.allowas_in = int(allowas_in)
                api.node_config(ndata,config_name)
            # Handle default_originate
            default_originate = data.get_from_box(intf, 'bgp.default_originate')
            if default_originate:
                # Report the parameter on the BGP session, based on ifindex
                for neigh in ndata.get('bgp', {}).get('neighbors', []):
                    if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                        neigh.default_originate = default_originate
                # Report the parameter on the BGP session (for VRF), based on ifindex
                if 'vrf' in intf:
                    for neigh in ndata.vrfs[intf.vrf].get('bgp', {}).get('neighbors', []):
                        if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                            neigh.default_originate = default_originate
                api.node_config(ndata,config_name)
            # Handle bgp_password
            bgp_password = data.get_from_box(intf, 'bgp.password')
            if bgp_password:
                # Report the parameter on the BGP session, based on ifindex
                for neigh in ndata.get('bgp', {}).get('neighbors', []):
                    if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                        neigh.password = bgp_password
                # Report the parameter on the BGP session (for VRF), based on ifindex
                if 'vrf' in intf:
                    for neigh in ndata.vrfs[intf.vrf].get('bgp', {}).get('neighbors', []):
                        if neigh.ifindex == intf.ifindex and neigh.type == 'ebgp':
                            neigh.password = bgp_password
                api.node_config(ndata,config_name)



#
# Common routines for create-topology script
#
import sys
import typing
import warnings
import argparse
import os
import textwrap
import pathlib

from jinja2 import Environment, PackageLoader, FileSystemLoader, StrictUndefined, make_logging_undefined
from box import Box,BoxList
from .data.global_vars import get_topology

from .utils.log import LOGGING, VERBOSE, DEBUG, QUIET, RAISE_ON_ERROR, WARNING
from .utils.log import MissingValue, IncorrectAttr, IncorrectValue, IncorrectType, FatalError, ErrorAbort
from .utils.log import fatal, error, exit_on_error, set_logging_flags, set_flag, print_verbose, debug_active
from .utils.strings import extra_data_printout,format_structured_dict,print_structured_dict

AF_LIST = ['ipv4','ipv6']
BGP_SESSIONS = ['ibgp','ebgp']

netsim_package_path = os.path.abspath(os.path.dirname(__file__))

#
# File functions
#

def open_output_file(fname: str) -> typing.TextIO:
  if fname == '-':
    return sys.stdout

  return open(fname,mode='w')

def close_output_file(f: typing.TextIO) -> None:
  if f.name != '<stdout>':
    f.close()

ruamel_attrs: typing.Final[dict] = {'version': (1,1)}

def get_yaml_string(x : typing.Any) -> str:
  global ruamel_attrs
  if isinstance(x, Box) or isinstance(x,BoxList):
    return x.to_yaml(ruamel_attrs=ruamel_attrs)
  if isinstance(x,dict):
    return Box(x).to_yaml(ruamel_attrs=ruamel_attrs)
  elif isinstance(x,list):
    return BoxList(x).to_yaml(ruamel_attrs=ruamel_attrs)
  else:
    return str(x)

#
# Find path to the module directory (needed for various templates)
#
def get_moddir() -> pathlib.Path:
  return pathlib.Path(__file__).resolve().parent

#
# Find a file in a search path
#
def find_file(path: str, search_path: typing.List[str]) -> typing.Optional[str]:
  for dirname in search_path:
    candidate = os.path.join(dirname, path)
    if os.path.exists(candidate):
      return candidate

  return None

def template(j2: str , data: typing.Dict, path: str, user_template_path: typing.Optional[str] = None) -> str:
  if path [0] in ('.','/'):                             # Absolute path or path relative to current directory?
    template_path = [ path ]
  else:                                                 # Path relative to netsim module, add module path to it
    template_path = [ str(get_moddir()) + "/" + path ]
  if not user_template_path is None:
    template_path = [ './' + user_template_path, os.path.expanduser('~/.netlab/'+user_template_path) ] + template_path
  if debug_active('template'):
    print(f"TEMPLATE PATH for {j2}: {template_path}")
  ENV = Environment(loader=FileSystemLoader(template_path), \
          trim_blocks=True,lstrip_blocks=True, \
          undefined=make_logging_undefined(base=StrictUndefined))
  template = ENV.get_template(j2)
  return template.render(**data)

#
# write_template: Applies a custom template (in_folder/j2) and writes it to the given file path (out_folder/filename)
#
def write_template(in_folder: str, j2: str, data: typing.Dict, out_folder: str, filename: str) -> None:
  if debug_active('template'):
    print(f"write_template {in_folder}/{j2} -> {out_folder}/{filename}")
  pathlib.Path(out_folder).mkdir(parents=True, exist_ok=True)
  out_file = f"{out_folder}/{filename}"
  with open(out_file,mode='w') as output:
    output.write(template(j2,data,in_folder))



#!/usr/bin/env python3
#
# Create expanded topology file, Ansible inventory, host vars, or Vagrantfile from
# topology file
#

import sys
import os
import glob
import pathlib
import pytest
import difflib
from box import Box

import utils

from netsim import common
from netsim import read_topology
from netsim import augment
from netsim.outputs import _TopologyOutput,ansible

def run_test(fname,local_defaults="topology-defaults.yml",sys_defaults="package:topology-defaults.yml"):
  topology = read_topology.load(fname,local_defaults,sys_defaults)
  common.exit_on_error()
  augment.main.transform(topology)
  common.exit_on_error()
  return topology

@pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
def test_transformation_cases(tmpdir):
  print("Starting transformation test cases")
  for test_case in list(glob.glob('topology/input/*yml')):
    print("Test case: %s" % test_case)
    topology = run_test(test_case)

    if topology.defaults.get("inventory"):
      print("Writing inventory... %s" % topology.defaults.inventory)
      ansible.ansible_inventory(topology,tmpdir+"/extra/hosts.yml",topology.defaults.get("inventory").replace("dump",""))
      ansible.ansible_config(tmpdir+"/ansible.cfg",tmpdir+"/hosts.yml")
      if topology.defaults.inventory == "dump":
        ansible.dump(topology)

    if topology.defaults.get("Output"):
      for output_format in topology.defaults.get("Output"):
        output_module = _TopologyOutput.load(output_format,topology.defaults.outputs[output_format])
        if output_module:
          output_module.write(Box(topology))
        else:
          common.error('Unknown output format %s' % output_format,common.IncorrectValue,'create')

    result = utils.transformation_results_yaml(topology)
    exp_test_case = "topology/expected/"+os.path.basename(test_case)
    expected = pathlib.Path(exp_test_case).read_text()
    if result != expected:
      print("Test case: %s FAILED" % test_case)
      sys.stdout.writelines(
        difflib.unified_diff(
          expected.splitlines(keepends=True),
          result.splitlines(keepends=True),
          fromfile='expected',tofile='result'))

    assert result == expected
    print("... succeeded, string length = %d" % len(result))

# Verbose test cases are executed only when we're doing a coverage report
#
def test_verbose_cases(tmpdir):
  if not sys.gettrace():
    return
  common.set_verbose()
  test_transformation_cases(tmpdir)

@pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
def test_error_cases():
  print("Starting error test cases")
  common.set_flag(raise_error = True)
  for test_case in list(glob.glob('errors/*yml')):
    print("Test case: %s" % test_case)
    common.err_count = 0
    with pytest.raises(common.ErrorAbort):
      topo = run_test(test_case)

"""
With the expanded default settings, it became impractical to run test cases
without the defaults. These test cases have been migrated into errors

@pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
def test_minimal_cases():
  print("Starting minimal (no-default) test cases")
  common.set_flag(raise_error = True)
  for test_case in list(glob.glob('minimal_errors/*yml')):
    print("Test case: %s" % test_case)
    common.err_count = 0
    with pytest.raises(common.ErrorAbort):
      run_test(test_case,None,None)
"""

if __name__ == "__main__":
  test_transformation_cases("/tmp")
#  test_error_cases()
#  test_minimal_cases()



#
# First-hop gateway transformation module
#
import typing
from box import Box

from . import _Module,get_effective_module_attribute
from .. import common
from .. import data
from ..augment.nodes import reserve_id
from ..augment import devices
from ..data.validate import validate_attributes,must_be_string

def check_gw_protocol(gw: Box, path: str, topology: Box) -> typing.Any:
  return must_be_string(
      parent=gw,
      key='protocol',
      path=path,
      module='gateway',
      valid_values=topology.defaults.gateway.attributes.protocols
    )

#
# Check whether a node supports all FHRP protocols configured on it
#

def check_protocol_support(node: Box, topology: Box) -> bool:
  features = devices.get_device_features(node,topology.defaults)
  proto_list = []
  OK = True

  for intf in node.interfaces:                                        # Iterate over interfaces
    if not 'gateway' in intf:                                         # ... and skip interfaces without FHRP
      continue
    gw_proto = intf.gateway.protocol                                  # We're asuming someone else did a sanity check on this value
    if gw_proto in proto_list:                                        # Already checked?
      continue

    proto_list.append(gw_proto)
    if not gw_proto in features.gateway.protocol:
      OK = False
      common.error(
        f'Node {node.name} ({node.device}) does not support gateway protocol {gw_proto}',
        common.IncorrectValue,
        'gateway')

  return OK

#
# Remove unicast IPv4 addresses from interfaces that use 'anycast' gateway protocol
# and have 'gateway.anycast.unicast' set to False
#
def cleanup_unicast_ip(node: Box) -> None:
  for intf in node.interfaces:
    if not intf.get('gateway',False):                               # This interface not using FHRP or FHRP is disabled
      continue

    if intf.gateway.protocol != 'anycast':                          # Leave non-anycast FHRP implementations alone, they need node IP addresses
      continue

    if data.get_from_box(intf,'gateway.anycast.unicast') is False:  # Are we forbidden to use unicast IP addresses together with anycast ones?
      intf.pop('ipv4',None)                                         # No unicast with anycast ==> pop the address

#
# Default settings copied onto individual links have parameters for every known FHRP protocol.
# We don't need those parameters on every interface -- this function cleans up unusud gateway protocol
# parameters from interfaces and returns a list of active protocols so we know what to clean on the
# node level.

def cleanup_protocol_parameters(node: Box,topology: Box) -> list:
  active_proto: list = []

  proto_list = topology.defaults.gateway.attributes.protocols         # List of known FHRP protocols
  for intf in node.interfaces:                                        # Iterate over interfaces
    if not 'gateway' in intf:                                         # ... and skip interfaces without FHRP
      continue

    gw_proto = intf.gateway.protocol                                  # We're asuming someone else did a sanity check on this value
    if not gw_proto in active_proto:                                  # Add the current protocol to the list of active protocols
      active_proto.append(gw_proto)

    for k in list(intf.gateway):                                      # Now iterate over all keywords in interface gateway settings
      if k != gw_proto and k in proto_list:                           # ... found FHRP protocol that is NOT the active protocol
        intf.gateway.pop(k,None)                                      # ... useless, pop it

  return active_proto

class FHRP(_Module):

  def module_init(self, topology: Box) -> None:
    gw = data.get_global_settings(topology,'gateway')
    if not gw or gw is None:
      common.error(
        f'Global/default gateway parameters are missing. We need at least a gateway ID',
        common.IncorrectType,
        'gateway')
      return

    check_gw_protocol(gw,'topology.gateway',topology)

    if not data.is_true_int(gw.id):
      common.error(
        f'Global/default gateway.id parameter is missing or not integer',
        common.IncorrectType,
        'gateway')
      return

    if gw.id > 0:
      reserve_id(gw.id)

  def link_pre_link_transform(self, link: Box, topology: Box) -> None:
    if not 'gateway' in link:
      return

    global_gateway = data.get_global_settings(topology,'gateway')
    if not global_gateway:                      # pragma: no cover
      return                                    # We know (from init) that we have global parameters. This is just to keep mypy happy

    if link.gateway is True:                    # We just want to do FHRP on the link ==> take global parameters
      link.gateway = global_gateway
    elif link.gateway is False:                 # We DEFINITELY don't want FHRP on this link ==> remove it and move on
      link.pop('gateway',None)
      return
    else:                                       # Otherwise merge global defaults with links settings (because we usually don't do that)
      check_gw_protocol(link.gateway,f'{link._linkname}',topology)
      link.gateway = global_gateway + link.gateway

    for k in ('id','protocol'):
      if not k in link.gateway or not link.gateway[k]:
        common.error(
          f'Gateway attribute {k} is missing in {link._linkname}\n' + \
          common.extra_data_printout(common.format_structured_dict(link)),
          common.MissingValue,
          'gateway')
        return

    if not data.is_true_int(link.gateway.id):
      common.error(
        f'Gateway.id parameter in {link._linkname} must be an integer\n' + \
          common.extra_data_printout(common.format_structured_dict(link)),
        common.IncorrectType,
        'gateway')
      return

    if link.gateway.id == -1:
        common.error(
          f'Cannot use -1 as the gateway ID in {link._linkname} -- that would be the broadcast address\n' + \
          common.extra_data_printout(common.format_structured_dict(link)),
          common.IncorrectValue,
          'gateway')

  def node_post_transform(self, node: Box, topology: Box) -> None:
    if not check_protocol_support(node,topology):
      return

    cleanup_unicast_ip(node)
    active_proto = cleanup_protocol_parameters(node,topology)         # Cleanup interface parameters and get a list of active protocols
    if 'gateway' in node:
      for k in list(node.gateway):                                    # Iterate over node-level gateway parameters
        if not k in active_proto:                                     # Not a parameter for a FHRP active on this node?
          node.gateway.pop(k,None)                                    # ... zap it!



#
# netlab config command
#
# Deploy custom configuration template to network devices
#
import typing
import argparse
import os
import glob
import subprocess
import shutil

from box import Box
from pathlib import Path

from .. import common
from .. import utils
from .. import read_topology
from . import external_commands

#
# CLI parser for 'netlab config' command
#
def test_parse(args: typing.List[str], settings: Box) -> argparse.Namespace:
  choices = ( p for p in settings.providers.keys() if 'start' in settings.providers[p] )
  parser = argparse.ArgumentParser(
    prog='netlab test',
    description='Test virtual lab installation')
  parser.add_argument(
    '-w','--work-directory',
    dest='workdir',
    action='store',
    default='test',
    help='Working directory (default: test)')
  parser.add_argument(
    '-v','--verbose',
    dest='verbose',
    action='count',
    default=0,
    help='Verbose logging')
  parser.add_argument(
    dest='provider',
    action='store',
    choices=list(choices),
    help='Run tests for the specified provider')

  return parser.parse_args(args)

def copy_topology(args: argparse.Namespace) -> None:
  if args.verbose:
    print("Creating test topology file in %s" % args.workdir)
  os.mkdir(args.workdir)
  os.chdir(args.workdir)
  moddir = Path(__file__).resolve().parent.parent

  src_topo = moddir / "templates/tests" / (args.provider + ".yml")
  try:
    shutil.copy(src_topo,"topology.yml")
  except:
    common.fatal("Cannot copy test topology file %s into %s" % (src_topo,args.workdir))
  if args.verbose:
    print("... done, moving on\n")

def create_configs() -> None:
  external_commands.print_step(2,"creating configuration files")
  if not external_commands.run_command(["netlab","create"],True):
    common.fatal("netlab create failed, aborting...","netlab test")

def cleanup_working_directory(args: argparse.Namespace) -> None:
  if args.verbose:
    print("\nCleaning up -- removing %s and its contents" % args.workdir)
  os.chdir("..")
  try:
    shutil.rmtree(args.workdir)
  except:
    print("... cannot clean up working directory, trying to run 'sudo rm -fr'")
    try:
      subprocess.run(["sudo","rm","-fr",args.workdir],check=True)
      print("... that worked ;)")
    except:
      print("... rm -fr failed, please clean up %s manually" % args.workdir)
      return

  if args.verbose:
    print("... done, test completed\n")

def run(cli_args: typing.List[str]) -> None:
  settings = read_topology.read_yaml('package:topology-defaults.yml')
  if not cli_args:
    common.fatal("Specify the virtualization environment to test or use -h to get help","test")
    return

  if not settings:
    common.fatal("Cannot read the global defaults","test")
    return

  args = test_parse(cli_args,settings)
  if os.path.exists(args.workdir):
    common.fatal("Directory %s already exists, aborting" % args.workdir,"test")

  if args.verbose:
    utils.log.set_logging_flags(args)

  external_commands.run_probes(settings,args.provider,1)
  copy_topology(args)
  create_configs()
  if not external_commands.run_command('netlab up'):
    common.fatal('netlab up failed, aborting','test')
  elif not external_commands.run_command('netlab down'):
    common.fatal('netlab down failed','test')
#  external_commands.start_lab(settings,args.provider,3)
#  external_commands.deploy_configs(4)
#  external_commands.stop_lab(settings,args.provider,5)
  cleanup_working_directory(args)



#
# Create D2 graph file
#
import typing

import yaml
import os
from box import Box

from .. import common
from ..data import get_from_box,get_box
from ..data.validate import must_be_list
from . import _TopologyOutput

'''
Copy default settings into a D2 map converting Python dictionaries into
dotted-name format D2 is using
'''
def dump_d2_dict(f : typing.TextIO, data: dict, indent: str) -> None: 
  for k,v in data.items():
    if isinstance(v,dict):
      dump_d2_dict(f,v,f'{indent}{k}.')
    else:
      v = f'"{v}"' if isinstance(v,str) else v
      f.write(f"{indent}{k}: {v}\n")

'''
Copy named D2 default settings into D2 output file. Just a convenience
wrapper around dump_d2_dict
'''
def copy_d2_attr(f : typing.TextIO, dt: str, settings: Box, indent: str = '') -> None: 
  if not dt in settings:
    return

  dump_d2_dict(f,settings[dt],indent)

'''
Find D2-specific device type for a given node and copy device-specific attributes
into D2 object. Another convenience wrapper around copy_d2_attr and dump_d2_dict
'''
def d2_node_attr(f : typing.TextIO, n: Box, settings: Box, indent: str = '') -> None:
  d2_type = n.d2.type or ''
  copy_d2_attr(f,d2_type,settings,indent)

'''
Create a node in D2 graph and add a label and styling attributes to it

indent parameter is used to create indented definitions within containers
'''
def node_with_label(f : typing.TextIO, n: Box, settings: Box, indent: str = '') -> None:
  f.write(f'{indent}{n.d2.name} {{\n')
  node_ip_str = ""
  if settings.node_address_label:
    node_ip = n.loopback.ipv4 or n.loopback.ipv6
    if not node_ip and n.interfaces:
      node_ip = n.interfaces[0].ipv4 or n.interfaces[0].ipv6
    if node_ip:
      node_ip_str = f'\\n{node_ip}'

  f.write(f"  {indent}label: \"{n.name} [{n.device}]{node_ip_str}\"\n")
  d2_node_attr(f,n,settings,indent+'  ')
  f.write(f'{indent}}}\n')

'''
Similar to node-with-label, create a LAN segment node in the D2 graph. Node name is
the LAN bridge name, node label is its IPv4 or IPv6 prefix.
'''
def network_with_label(f : typing.TextIO, n: Box, settings: Box, indent: str = '') -> None:
  f.write(f'{indent}{n.bridge}' + '{\n')
  f.write(f'  label: {n.prefix.ipv4 or n.prefix.ipv6 or n.bridge}\n')
  copy_d2_attr(f,'lan',settings,'  '+indent)
  f.write('}\n')
#  f.write('style=filled fillcolor="%s" fontsize=11' % (settings.colors.get("stub","#d1bfab")))

'''
Add an arrowhead label to a connection
'''
def edge_label(f : typing.TextIO, direction: str, data: Box, subnet: bool = True) -> None:
  addr = data.ipv4 or data.ipv6
  if isinstance(addr,str):
    if not subnet:
      addr = addr.split('/')[0]
    f.write(f"  {direction}-arrowhead.label: '{addr}'\n")

'''
Create a P2P connection between two nodes
'''
def edge_p2p(f : typing.TextIO, l: Box, labels: typing.Optional[bool] = False) -> None:
  f.write(f"{l.interfaces[0].node} -- {l.interfaces[1].node} {{\n")
  if labels:
    edge_label(f,'source',l.interfaces[0],True)
    edge_label(f,'target',l.interfaces[1],True)
  f.write("}\n")

'''
Create a connection between a node and a LAN segment
'''
def edge_node_net(f : typing.TextIO, link: Box, ifdata: Box, labels: typing.Optional[bool] = False) -> None:
  f.write(f"{ifdata.node} -> {link.bridge} {{\n")
  if labels:
    edge_label(f,'source',ifdata,False)
  f.write("}\n")

'''
Create an ASN or group container
'''
def as_start(f : typing.TextIO, asn: str, label: typing.Optional[str], settings: Box) -> None:
  f.write(f'{asn} {{\n')
  copy_d2_attr(f,'container',settings,'-  ')
  asn = asn.replace('_',' ')
  f.write('  label: '+ (f'{label} ({asn})' if label else asn)+'\n')

'''
Create a topology graph as a set of containers (groups or ASNs)
'''
def graph_clusters(f : typing.TextIO, clusters: Box, settings: Box, nodes: Box) -> None:
  for asn in clusters.keys():
    as_start(f,asn,clusters[asn].get('name',None),settings)
    for n in clusters[asn].nodes.values():
      node_with_label(f,n,settings,'  ')                    # Create a node within a cluster
      nodes[n.name].d2.name = f'{asn}.{n.name}'             # And change the D2 node name that will be used in connections
    f.write('}\n')

def build_maps(topology: Box) -> Box:
  maps = Box({},default_box=True,box_dots=True)
  for name,n in topology.nodes.items():
    maps.nodes[name] = n

  if 'bgp' in topology.get('module',[]):
    for name,n in topology.nodes.items():
      bgp_as = get_from_box(n,'bgp.as')
      if bgp_as:
        bgp_as = f'AS_{bgp_as}'
        maps.bgp[bgp_as].nodes[n.name] = n

  if 'bgp' in topology and 'as_list' in topology.bgp:
    for (asn,asdata) in topology.bgp.as_list.items():
      if 'name' in asdata and asn in maps.bgp:
        maps.bgp[asn].name = asdata.name

  return maps

'''
add_groups -- use topology groups as graph clustering mechanism
'''
def add_groups(maps: Box, groups: list, topology: Box) -> None:
  if not 'groups' in topology:
    return

  placed_hosts = []
  for g,v in topology.groups.items():
    if g in groups:
      for n in v.members:
        if n in placed_hosts:
          common.error(
            f'Cannot create overlapping graph clusters: node {n} is in two groups',
            common.IncorrectValue,
            'graph')
          continue
        else:
          maps.groups[g].nodes[n] = topology.nodes[n]
          placed_hosts.append(n)

def graph_topology(topology: Box, fname: str, settings: Box,g_format: typing.Optional[list]) -> bool:
  f = common.open_output_file(fname)
  maps = build_maps(topology)

  if 'groups' in settings:
    must_be_list(
      parent=settings,
      key='groups',path='defaults.outputs.graph',
      true_value=list(topology.get('groups',{}).keys()),
      create_empty=True,
      module='graph')
    add_groups(maps,settings.groups,topology)
    graph_clusters(f,maps.groups,settings,topology.nodes)
  elif 'bgp' in maps and settings.as_clusters:
    graph_clusters(f,maps.bgp,settings,topology.nodes)
  else:
    for name,n in topology.nodes.items():
      node_with_label(f,n,settings)

  for l in topology.links:
    for intf in l.interfaces:
      intf.node = topology.nodes[intf.node].d2.name
    if l.type == "p2p":
      edge_p2p(f,l,settings.interface_labels)
    else:
      if not l.bridge:
        common.error('Found a lan/stub link without a bridge name, skipping',common.IncorrectValue,'graph')
        next
      network_with_label(f,l,settings)
      for ifdata in l.interfaces:
        if ifdata.node in maps.nodes:
          edge_node_net(f,l,ifdata,settings.interface_labels)

  f.close()
  return True

'''
Create a BGP session as a connection between two nodes

* Select the connection (arrow) type based on whether a connection is a RR-to-client session
* Copy IBGP or EBGP attributes to the connection

Please note that we call this function once for every pair of nodes, so it has to deal with
RR being the first (node) or the second (session) connection endpoint.
'''
def bgp_session(f : typing.TextIO, node: Box, session: Box, settings: Box, rr_session: bool, nodes: Box) -> None:
  arrow_dir = '<->'
  if rr_session:
    if session.type == 'ibgp':
      if 'rr' in node.bgp and node.bgp.rr and not 'rr' in session:
        arrow_dir = '->'
      if not 'rr' in node.bgp and 'rr' in session:
        arrow_dir = '<-'

  f.write(f'{node.d2.name} {arrow_dir} {nodes[session.name].d2.name} {{\n')
  copy_d2_attr(f,session.type,settings,'  ')
  f.write('}\n')

def graph_bgp(topology: Box, fname: str, settings: Box,g_format: typing.Optional[list]) -> bool:
  if not 'bgp' in topology.get('module',{}):
    common.error('BGP graph format can only be used to draw topologies using BGP')
    return False

  f = common.open_output_file(fname)

  rr_session = settings.get('rr_sessions',False)
  if g_format is not None and len(g_format) > 1:
    rr_session = g_format[1] == 'rr'

  maps = build_maps(topology)
  graph_clusters(f,maps.bgp,settings,topology.nodes)

  for name,n in topology.nodes.items():
    if 'bgp' in n:
      for neighbor in n.bgp.get('neighbors',[]):
        if neighbor.name > n.name:
          bgp_session(f,n,neighbor,settings,rr_session,topology.nodes)

  f.close()
  return True

graph_dispatch = {
  'topology': graph_topology,
  'bgp': graph_bgp
}

'''
Set node attributes needed by D2 output module:

* D2 shape name used in connections -- set to node name, can be modified to include container name
* D2 shape type -- used to copy D2 style attributes from system defaults to D2 graph file
'''
def set_d2_attr(topology: Box) -> None:
  for n,ndata in topology.nodes.items():
    dev_data = topology.defaults.devices[ndata.device]
    ndata.d2.type = dev_data.graphite.icon or 'router'
    ndata.d2.name = n

class Graph(_TopologyOutput):

  def write(self, topology: Box) -> None:
    graphfile = self.settings.filename or 'graph.d2'
    output_format = 'topology'

    topology = get_box(topology)                       # Create a local copy of the topology
    set_d2_attr(topology)
    if hasattr(self,'filenames'):
      graphfile = self.filenames[0]
      if len(self.filenames) > 1:
        common.error('Extra output filename(s) ignored: %s' % str(self.filenames[1:]),common.IncorrectValue,'graph')

    if self.format:
      output_format = self.format[0]

    if output_format in graph_dispatch:
      if graph_dispatch[output_format](topology,graphfile,self.settings,self.format):
        print("Created graph file %s in %s format" % (graphfile, output_format))
    else:
      formats = ', '.join(graph_dispatch.keys())
      common.error('Unknown graph format, use one of %s' % formats,common.IncorrectValue,'graph')



import typing, netaddr
from box import Box
from netsim import common
from netsim import utils
from netsim import api
from netsim import data

"""
Depth-first evaluation of changed parameters:

* Iterate over the dictionary of changes
* Recurse if the value is a dictionary
* Evaluate a string value as formatted string
* Skip everything else (until someone figures out we need more)
"""
def eval_changed_parameters(change: Box, ctx_data: Box) -> None:
  for k in list(change.keys()):
    if isinstance(change[k],Box):
      eval_changed_parameters(change[k],ctx_data)
    elif isinstance(change[k],str):
      if '{' in change[k]:
        change[k] = utils.strings.eval_format(change[k],ctx_data)

"""
Merge changed parameters evaluated by multilab into the lab topology

The changes have to be made 'in place' as we cannot return new topology from
plugin initialization code.
"""
def merge_changes(topology: Box, change: Box) -> None:
  for k in list(change.keys()):                                       # Iterate over changed parameters
    if isinstance(change[k],Box) and isinstance(topology[k],Box):     # Merging a hierarchical data structure into topology dictionary?
      topology[k] = topology[k] + change[k]                           # ... easy task, Box addition saves the day
    elif isinstance(change[k],list) and isinstance(topology[k],list): # We can add the changed list to the original list
      topology[k].extend(change[k])
    else:                                                             # Otherwise overwrite the original value. Bad luck.
      topology[k] = change[k]

"""
Main multilab code:

* Validate default settings
* Evaluate parameters to be changed
* Merge changed parameters into topology
"""
def init(topology: Box) -> None:
  mlab = topology.defaults.multilab
  abort = False
  for kw in ['id','change']:                                          # Check that we have all default parameters needed for multilab to work
    if not kw in mlab:
      common.error(f'multilab plugin requires defaults.multilab.{kw} parameter',common.MissingValue,'multilab')
      abort = True

  if abort:
    return

  data.types.must_be_int(                                             # Now validate that multilab.id is an integer less than 200
    parent=mlab,
    key='id',
    path='defaults.multilab',
    module='multilab',
    min_value=1,
    max_value=200)

  data.types.must_be_dict(                                            # ... and that multilab.change is a dictionary
    parent=mlab,
    key='change',
    path='defaults.multilab',
    module='multilab')

  ctx_data = data.get_box(topology)
  ctx_data.id = mlab.id
  eval_changed_parameters(mlab.change,ctx_data)                       # Evaluate changed parameters
  if common.debug_active('plugin'):                                   # Print the results if we're debugging
    print(f'MULTILAB CHANGES\n==============\n{mlab.change.to_yaml()}')

  merge_changes(topology,mlab.change)                                 # And merge the changes with the topology
